"""Command models to configure the stored labware pool of a Flex Stacker.."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Type

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from ...errors import (
    ErrorOccurrence,
)
from ...errors.exceptions import FlexStackerNotLogicallyEmptyError
from ...types import (
    SYSTEM_LOCATION,
    InStackerHopperLocation,
    LabwareLocationSequence,
    NotOnDeckLocationSequenceComponent,
    StackerStoredLabwareGroup,
)
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from .common import (
    INITIAL_COUNT_DESCRIPTION,
    INITIAL_STORED_LABWARE_DESCRIPTION,
    adapter_location_sequences_with_default,
    build_ids_to_fill,
    build_or_assign_labware_to_hopper,
    labware_location_base_sequence,
    labware_locations_for_group,
    lid_location_sequences_with_default,
    primary_location_sequences,
    validate_labware_pool_compatible_with_stacker,
)

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import EquipmentHandler
    from opentrons.protocol_engine.resources import ModelUtils
    from opentrons.protocol_engine.state.state import StateView

SetStoredLabwareCommandType = Literal["flexStacker/setStoredLabware"]


class StackerStoredLabwareDetails(BaseModel):
    """The parameters defining a labware to be stored in the stacker."""

    loadName: str = Field(
        ..., description="Name used to reference the definition of this labware."
    )
    namespace: str = Field(
        ..., description="Namespace of the definition of this labware."
    )
    version: int = Field(..., description="Version of the definition of this labware.")


class SetStoredLabwareParams(BaseModel):
    """Input parameters for a setStoredLabware command."""

    moduleId: str = Field(
        ...,
        description="Unique ID of the Flex Stacker.",
    )
    primaryLabware: StackerStoredLabwareDetails = Field(
        ...,
        description="The details of the primary labware (i.e. not the lid or adapter, if any) stored in the stacker.",
    )
    lidLabware: StackerStoredLabwareDetails | SkipJsonSchema[None] = Field(
        default=None,
        description="The details of the lid on the primary labware, if any.",
    )
    adapterLabware: StackerStoredLabwareDetails | SkipJsonSchema[None] = Field(
        default=None,
        description="The details of the adapter under the primary labware, if any.",
    )
    initialCount: Optional[Annotated[int, Field(ge=0)]] = Field(
        None,
        description=INITIAL_COUNT_DESCRIPTION,
    )
    initialStoredLabware: list[StackerStoredLabwareGroup] | None = Field(
        None,
        description=INITIAL_STORED_LABWARE_DESCRIPTION,
    )
    poolOverlapOverride: Optional[float] = Field(
        None,
        description=(
            "Override for the Z stacking overlap of the labware pool. If not "
            "provided, the protocol engine will calculate the overlap based on "
            "the stacking offsets provided in the labware definitions."
        ),
    )


class SetStoredLabwareResult(BaseModel):
    """Result data from a setstoredlabware command."""

    primaryLabwareDefinition: LabwareDefinition = Field(
        ..., description="The definition of the primary labware."
    )
    lidLabwareDefinition: LabwareDefinition | SkipJsonSchema[None] = Field(
        None, description="The definition of the lid on the primary labware, if any."
    )
    adapterLabwareDefinition: LabwareDefinition | SkipJsonSchema[None] = Field(
        None,
        description="The definition of the adapter under the primary labware, if any.",
    )
    storedLabware: list[StackerStoredLabwareGroup] = Field(
        ..., description="The primary labware loaded into the stacker labware pool."
    )
    count: int = Field(
        ..., description="The number of labware now stored in the hopper."
    )
    originalPrimaryLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The previous position of each primary labware, in the same order as storedLabware.",
    )
    originalAdapterLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The previous position of each adapter labware, in the same order as storedLabware. None if the pool does not specify an adapter.",
    )
    originalLidLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The previous position of each lid labware, in the same order as storedLabware. None if the pool does not specify a lid.",
    )
    newPrimaryLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The new position of each primary labware, in the same order as storedLabware.",
    )
    newAdapterLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The new position of each adapter labware, in the same order as storedLabware. None if the pool does not specify an adapter.",
    )
    newLidLabwareLocationSequences: (
        list[LabwareLocationSequence] | SkipJsonSchema[None]
    ) = Field(
        None,
        description="The new position of each lid labware, in the same order as storedLabware. None if the pool does not specify a lid labware.",
    )


class SetStoredLabwareImpl(
    AbstractCommandImpl[SetStoredLabwareParams, SuccessData[SetStoredLabwareResult]]
):
    """Implementation of a setstoredlabware command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._model_utils = model_utils

    async def execute(
        self, params: SetStoredLabwareParams
    ) -> SuccessData[SetStoredLabwareResult]:
        """Execute the setstoredlabwarecommand."""
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )
        location = self._state_view.modules.get_location(params.moduleId)

        if len(stacker_state.contained_labware_bottom_first) != 0:
            # Note: this error catches if the protocol tells us the stacker is not empty, making this command
            # invalid at this point in the protocol. This error is not recoverable and should occur during
            # analysis; the protocol must be changed.

            raise FlexStackerNotLogicallyEmptyError(
                message=(
                    f"The Flex Stacker in {location} must be known to be empty before reconfiguring its labware pool, but it has "
                    f"{len(stacker_state.contained_labware_bottom_first)} labware"
                )
            )

        labware_def, _ = await self._equipment.load_definition_for_details(
            load_name=params.primaryLabware.loadName,
            namespace=params.primaryLabware.namespace,
            version=params.primaryLabware.version,
        )
        lid_def: LabwareDefinition | None = None
        if params.lidLabware:
            lid_def, _ = await self._equipment.load_definition_for_details(
                load_name=params.lidLabware.loadName,
                namespace=params.lidLabware.namespace,
                version=params.lidLabware.version,
            )
        adapter_def: LabwareDefinition | None = None
        if params.adapterLabware:
            adapter_def, _ = await self._equipment.load_definition_for_details(
                load_name=params.adapterLabware.loadName,
                namespace=params.adapterLabware.namespace,
                version=params.adapterLabware.version,
            )

        pool_definitions = (
            self._state_view.labware.stacker_labware_pool_to_ordered_list(
                labware_def, lid_def, adapter_def
            )
        )

        validate_labware_pool_compatible_with_stacker(
            pool_primary_definition=labware_def,
            pool_adapter_definition=adapter_def,
            pool_lid_definition=lid_def,
        )

        pool_height = self._state_view.geometry.get_height_of_labware_stack(
            pool_definitions
        )

        pool_overlap = (
            params.poolOverlapOverride
            if params.poolOverlapOverride
            else self._state_view.labware.get_stacker_labware_overlap_offset(
                pool_definitions
            ).z
        )

        max_pool_count = self._state_view.modules.stacker_max_pool_count_by_height(
            params.moduleId,
            pool_height,
            pool_overlap,
        )
        groups_to_load = build_ids_to_fill(
            params.adapterLabware is not None,
            params.lidLabware is not None,
            params.initialStoredLabware,
            params.initialCount,
            max_pool_count,
            0,
            self._model_utils,
        )
        state_update, new_contained_labware = await build_or_assign_labware_to_hopper(
            labware_def,
            adapter_def,
            lid_def,
            params.moduleId,
            groups_to_load,
            [],
            self._equipment,
            self._state_view,
        )

        state_update = state_update.update_flex_stacker_labware_pool_definition(
            params.moduleId,
            max_pool_count,
            pool_overlap,
            pool_height,
            labware_def,
            adapter_def,
            lid_def,
        )
        original_location_sequences = [
            labware_locations_for_group(
                group,
                labware_location_base_sequence(
                    group,
                    self._state_view,
                    [
                        NotOnDeckLocationSequenceComponent(
                            logicalLocationName=SYSTEM_LOCATION
                        )
                    ],
                ),
            )
            for group in new_contained_labware
        ]
        new_location_sequences = [
            labware_locations_for_group(
                group, [InStackerHopperLocation(moduleId=params.moduleId)]
            )
            for group in new_contained_labware
        ]
        return SuccessData(
            public=SetStoredLabwareResult.model_construct(
                primaryLabwareDefinition=labware_def,
                lidLabwareDefinition=lid_def,
                adapterLabwareDefinition=adapter_def,
                count=len(new_contained_labware),
                storedLabware=new_contained_labware,
                originalPrimaryLabwareLocationSequences=primary_location_sequences(
                    original_location_sequences
                ),
                originalAdapterLabwareLocationSequences=adapter_location_sequences_with_default(
                    original_location_sequences, adapter_def
                ),
                originalLidLabwareLocationSequences=lid_location_sequences_with_default(
                    original_location_sequences, lid_def
                ),
                newPrimaryLabwareLocationSequences=primary_location_sequences(
                    new_location_sequences
                ),
                newAdapterLabwareLocationSequences=adapter_location_sequences_with_default(
                    new_location_sequences, adapter_def
                ),
                newLidLabwareLocationSequences=lid_location_sequences_with_default(
                    new_location_sequences, lid_def
                ),
            ),
            state_update=state_update,
        )


class SetStoredLabware(
    BaseCommand[SetStoredLabwareParams, SetStoredLabwareResult, ErrorOccurrence]
):
    """A command to setstoredlabware the Flex Stacker."""

    commandType: SetStoredLabwareCommandType = "flexStacker/setStoredLabware"
    params: SetStoredLabwareParams
    result: Optional[SetStoredLabwareResult] = None

    _ImplementationCls: Type[SetStoredLabwareImpl] = SetStoredLabwareImpl


class SetStoredLabwareCreate(BaseCommandCreate[SetStoredLabwareParams]):
    """A request to execute a Flex Stacker SetStoredLabware command."""

    commandType: SetStoredLabwareCommandType = "flexStacker/setStoredLabware"
    params: SetStoredLabwareParams

    _CommandCls: Type[SetStoredLabware] = SetStoredLabware



"""Test wait for tasks."""

from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons.protocol_engine.actions import ActionDispatcher
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.wait_for_tasks import (
    WaitForTasksImplementation,
    WaitForTasksParams,
    WaitForTasksResult,
)
from opentrons.protocol_engine.errors import ErrorOccurrence
from opentrons.protocol_engine.errors.exceptions import TaskFailedError
from opentrons.protocol_engine.execution import RunControlHandler
from opentrons.protocol_engine.execution.task_handler import TaskHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types.tasks import FinishedTask, Task


async def test_wait_for_tasks_implementation_no_error(
    decoy: Decoy,
    run_control: RunControlHandler,
    real_task_handler: TaskHandler,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    state_view: StateView,
) -> None:
    """It should wait for a list of tasks to complete using the RunControlHandler.

    No tasks have errors so NO exception should be raised.
    """
    subject = WaitForTasksImplementation(
        run_control=run_control, task_handler=real_task_handler, state_view=state_view
    )
    # Tasks to wait for
    task_ids = ["task1", "task2"]
    data = WaitForTasksParams(task_ids=task_ids)

    decoy.when(state_view.tasks.get_failed_tasks(task_ids)).then_return([])
    result = await subject.execute(data)
    for task in task_ids:
        fake_task = decoy.mock(cls=Task)
        decoy.when(state_view.tasks.get(task)).then_return(fake_task)

    decoy.verify(await run_control.wait_for_tasks(task_ids))
    assert result == SuccessData(public=WaitForTasksResult(task_ids=task_ids))


async def test_wait_for_tasks_implementation_with_error(
    decoy: Decoy,
    run_control: RunControlHandler,
    real_task_handler: TaskHandler,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    state_view: StateView,
) -> None:
    """It should wait for a list of tasks to complete using the RunControlHandler.

    One task fails with a TaskFailedError so an exception SHOULD be raised.
    """
    subject = WaitForTasksImplementation(
        run_control=run_control, task_handler=real_task_handler, state_view=state_view
    )
    task_ids = ["task1", "task2"]
    data = WaitForTasksParams(task_ids=task_ids)
    created_timestamp = datetime.now()
    decoy.when(model_utils.get_timestamp()).then_return(created_timestamp)
    decoy.when(state_view.tasks.get_failed_tasks(task_ids)).then_return([task_ids[0]])
    decoy.when(state_view.tasks.get_finished(task_ids[0])).then_return(
        FinishedTask(
            id=task_ids[0],
            createdAt=matchers.Anything(),
            finishedAt=matchers.Anything(),
            error=ErrorOccurrence(
                id="error",
                createdAt=created_timestamp,
                errorType="TaskFailedError",
                detail="detail",
            ),
        )
    )

    with pytest.raises(TaskFailedError) as exc_info:
        await subject.execute(data)

    err = exc_info.value
    assert isinstance(err, TaskFailedError)
    assert err.message == "1 tasks failed."

    for task in task_ids:
        fake_task = decoy.mock(cls=Task)
        decoy.when(state_view.tasks.get(task)).then_return(fake_task)
    decoy.verify(await run_control.wait_for_tasks(task_ids))



"""Liquid state store tests.

DEPRECATED: Testing LiquidStore independently of LiquidView is no longer helpful.
Try to add new tests to test_liquid_state.py, where they can be tested together,
treating LiquidState as a private implementation detail.
"""

import pytest

from opentrons.protocol_engine import Liquid
from opentrons.protocol_engine.actions.actions import AddLiquidAction
from opentrons.protocol_engine.state.liquids import LiquidStore


@pytest.fixture
def subject() -> LiquidStore:
    """Liquid store test subject."""
    return LiquidStore()


def test_handles_add_liquid(subject: LiquidStore) -> None:
    """It should add the liquid to the state."""
    expected_liquid = Liquid(
        id="water-id", displayName="water", description="water-desc"
    )
    subject.handle_action(
        AddLiquidAction(
            Liquid(id="water-id", displayName="water", description="water-desc")
        )
    )

    assert len(subject.state.liquids_by_id) == 1

    assert subject.state.liquids_by_id["water-id"] == expected_liquid



"""Test the common utility functions used in transfers."""

from contextlib import nullcontext as does_not_raise
from typing import Any, ContextManager, Iterable, List, Tuple

import pytest

from opentrons.protocols.advanced_control.transfers.common import (
    Target,
    check_valid_volume_parameters,
    expand_for_volume_constraints,
    expand_for_volume_constraints_for_liquid_classes,
)


@pytest.mark.parametrize(
    argnames=["disposal_volume", "air_gap", "max_volume", "expected_raise"],
    argvalues=[
        (9.9, 9.9, 10, pytest.raises(ValueError, match="The sum of")),
        (9.9, 10, 10, pytest.raises(ValueError, match="The air gap must be less than")),
        (
            10,
            9.9,
            10,
            pytest.raises(ValueError, match="The disposal volume must be less than"),
        ),
        (9.9, 9.9, 20, does_not_raise()),
    ],
)
def test_check_valid_volume_parameters(
    disposal_volume: float,
    air_gap: float,
    max_volume: float,
    expected_raise: ContextManager[Any],
) -> None:
    """It should raise the expected error for invalid parameters."""
    with expected_raise:
        check_valid_volume_parameters(
            disposal_volume=disposal_volume,
            air_gap=air_gap,
            max_volume=max_volume,
        )


@pytest.mark.parametrize(
    argnames=["volumes", "targets", "max_volume", "expanded_list_result"],
    argvalues=[
        (
            [60, 70, 75],
            [(("a", "b"), (1, 2)), (("c", "d"), (3, 4)), (("e", "f"), (5, 6))],
            20,
            [
                (20, (("a", "b"), (1, 2))),
                (20, (("a", "b"), (1, 2))),
                (20, (("a", "b"), (1, 2))),
                (20, (("c", "d"), (3, 4))),
                (20, (("c", "d"), (3, 4))),
                (15, (("c", "d"), (3, 4))),
                (15, (("c", "d"), (3, 4))),
                (20, (("e", "f"), (5, 6))),
                (20, (("e", "f"), (5, 6))),
                (17.5, (("e", "f"), (5, 6))),
                (17.5, (("e", "f"), (5, 6))),
            ],
        ),
        (
            [100, 100, 100],
            ["dest1", "dest2", "dest3"],
            50,
            [
                (50, "dest1"),
                (50, "dest1"),
                (50, "dest2"),
                (50, "dest2"),
                (50, "dest3"),
                (50, "dest3"),
            ],
        ),
    ],
)
def test_expand_for_volume_constraints(
    volumes: Iterable[float],
    targets: Iterable[Target],
    max_volume: float,
    expanded_list_result: List[Tuple[float, Target]],
) -> None:
    """It should create a list of volume and target transfers, fitting the volumes within the max."""
    result = expand_for_volume_constraints(
        volumes=volumes,
        targets=targets,
        max_volume=max_volume,
    )
    assert list(result) == expanded_list_result


@pytest.mark.parametrize(
    argnames=["volumes", "targets", "max_volume", "air_gap", "expanded_list_result"],
    argvalues=[
        (
            [25, 25],
            ["dest1", "dest2"],
            50,
            24.0,
            [(25, "dest1"), (25, "dest2")],
        ),
        (
            [50, 50],
            ["dest1", "dest2"],
            50,
            0.0,
            [(50, "dest1"), (50, "dest2")],
        ),
        (
            [50, 50],
            ["dest1", "dest2"],
            50,
            5.0,
            [(25, "dest1"), (25, "dest1"), (25, "dest2"), (25, "dest2")],
        ),
        (
            [75, 75],
            ["dest1", "dest2"],
            50,
            5.0,
            [(37.5, "dest1"), (37.5, "dest1"), (37.5, "dest2"), (37.5, "dest2")],
        ),
        (
            [100, 100],
            ["dest1", "dest2"],
            50,
            0.0,
            [(50, "dest1"), (50, "dest1"), (50, "dest2"), (50, "dest2")],
        ),
        (
            [100, 100],
            ["dest1", "dest2"],
            50,
            5.0,
            [
                (100 / 3, "dest1"),
                (100 / 3, "dest1"),
                (100 / 3, "dest1"),
                (100 / 3, "dest2"),
                (100 / 3, "dest2"),
                (100 / 3, "dest2"),
            ],
        ),
        (
            [103, 103],
            ["dest1", "dest2"],
            50,
            0.0,
            [
                (103 / 3, "dest1"),
                (103 / 3, "dest1"),
                (103 / 3, "dest1"),
                (103 / 3, "dest2"),
                (103 / 3, "dest2"),
                (103 / 3, "dest2"),
            ],
        ),
    ],
)
def test_expand_for_volume_constraints_for_liquid_classes(
    volumes: Iterable[float],
    targets: Iterable[Target],
    max_volume: float,
    air_gap: float,
    expanded_list_result: List[Tuple[float, Target]],
) -> None:
    """It should create a list of volume and target transfers, splitting volumes equally if required."""
    result = expand_for_volume_constraints_for_liquid_classes(
        volumes=volumes,
        targets=targets,
        max_volume=max_volume,
        air_gap=air_gap,
    )
    assert list(result) == expanded_list_result



"""Tests for the module state store handling flex stacker state."""

from typing import Dict, Optional, Set, cast
from unittest.mock import sentinel

import pytest

from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition,
)
from opentrons_shared_data.robot.types import RobotType

import opentrons.protocol_engine.errors as errors
from opentrons.protocol_engine import actions
from opentrons.protocol_engine.state.addressable_areas import (
    AddressableAreaState,
    AddressableAreaView,
)
from opentrons.protocol_engine.state.config import Config
from opentrons.protocol_engine.state.module_substates import (
    FlexStackerId,
    FlexStackerSubState,
)
from opentrons.protocol_engine.state.modules import ModuleStore, ModuleView
from opentrons.protocol_engine.types import (
    AddressableArea,
    DeckConfigurationType,
    DeckType,
    ModuleDefinition,
    PotentialCutoutFixture,
    StackerStoredLabwareGroup,
)


def get_addressable_area_view(
    loaded_addressable_areas_by_name: Optional[Dict[str, AddressableArea]] = None,
    potential_cutout_fixtures_by_cutout_id: Optional[
        Dict[str, Set[PotentialCutoutFixture]]
    ] = None,
    deck_definition: Optional[DeckDefinitionV5] = None,
    deck_configuration: Optional[DeckConfigurationType] = None,
    robot_type: RobotType = "OT-3 Standard",
    use_simulated_deck_config: bool = False,
) -> AddressableAreaView:
    """Get a labware view test subject."""
    state = AddressableAreaState(
        loaded_addressable_areas_by_name=loaded_addressable_areas_by_name or {},
        potential_cutout_fixtures_by_cutout_id=potential_cutout_fixtures_by_cutout_id
        or {},
        deck_definition=deck_definition or cast(DeckDefinitionV5, {"otId": "fake"}),
        deck_configuration=deck_configuration or [],
        robot_definition={
            "displayName": "OT-3",
            "robotType": "OT-3 Standard",
            "models": ["OT-3 Standard"],
            "extents": [477.2, 493.8, 0.0],
            "paddingOffsets": {
                "rear": -177.42,
                "front": 51.8,
                "leftSide": 31.88,
                "rightSide": -80.32,
            },
            "mountOffsets": {
                "left": [-13.5, -60.5, 255.675],
                "right": [40.5, -60.5, 255.675],
                "gripper": [84.55, -12.75, 93.85],
            },
        },
        robot_type=robot_type,
        use_simulated_deck_config=use_simulated_deck_config,
    )

    return AddressableAreaView(state=state)


@pytest.fixture
def ot3_state_config() -> Config:
    """Get a ProtocolEngine state config for the Flex."""
    return Config(
        robot_type="OT-3 Standard",
        deck_type=DeckType.OT3_STANDARD,
    )


@pytest.fixture
def subject(
    ot3_state_config: Config,
) -> ModuleStore:
    """Get a ModuleStore for the flex."""
    return ModuleStore(config=ot3_state_config, deck_fixed_labware=[])


@pytest.fixture
def module_view(subject: ModuleStore) -> ModuleView:
    """Get a ModuleView for the ModuleStore."""
    return ModuleView(state=subject._state)


def test_add_module_action(
    subject: ModuleStore,
    module_view: ModuleView,
    flex_stacker_v1_def: ModuleDefinition,
) -> None:
    """It should create a flex stacker substate."""
    action = actions.AddModuleAction(
        module_id="someModuleId",
        serial_number="someSerialNumber",
        definition=flex_stacker_v1_def,
        module_live_data={"status": "idle", "data": {}},
    )

    with pytest.raises(errors.ModuleNotLoadedError):
        module_view.get_flex_stacker_substate("someModuleId")

    subject.handle_action(action)

    result = module_view.get_flex_stacker_substate("someModuleId")

    assert result == FlexStackerSubState(
        module_id=FlexStackerId("someModuleId"),
        pool_primary_definition=None,
        pool_adapter_definition=None,
        pool_lid_definition=None,
        contained_labware_bottom_first=[],
        max_pool_count=0,
        pool_overlap=0,
        pool_height=0,
    )


@pytest.mark.parametrize(
    argnames=["primary_def", "lid_def", "adapter_def", "result"],
    argvalues=[
        pytest.param(
            sentinel.primary, None, None, [sentinel.primary], id="primary-only"
        ),
        pytest.param(
            sentinel.primary,
            sentinel.lid,
            None,
            [sentinel.lid, sentinel.primary],
            id="primary-and-lid",
        ),
        pytest.param(
            sentinel.primary,
            None,
            sentinel.adapter,
            [sentinel.primary, sentinel.adapter],
            id="primary-and-adapter",
        ),
        pytest.param(
            sentinel.primary,
            sentinel.lid,
            sentinel.adapter,
            [sentinel.lid, sentinel.primary, sentinel.adapter],
            id="primary-and-adapter-and-lid",
        ),
        pytest.param(None, None, None, None, id="none"),
    ],
)
def test_get_labware_definition_list(
    primary_def: LabwareDefinition | None,
    lid_def: LabwareDefinition | None,
    adapter_def: LabwareDefinition | None,
    result: list[LabwareDefinition] | None,
) -> None:
    """It should return definitions in proper order."""
    subject = FlexStackerSubState(
        module_id=FlexStackerId("someModuleId"),
        pool_primary_definition=primary_def,
        pool_adapter_definition=adapter_def,
        pool_lid_definition=lid_def,
        contained_labware_bottom_first=[],
        max_pool_count=5,
        pool_overlap=0,
        pool_height=0,
    )
    assert subject.get_pool_definition_ordered_list() == result


def test_get_contained_labware() -> None:
    """It should present a list of contained labware."""
    subject = FlexStackerSubState(
        module_id=FlexStackerId("someModuleId"),
        pool_primary_definition=sentinel.primary_def,
        pool_adapter_definition=sentinel.adapter_def,
        pool_lid_definition=sentinel.lid_def,
        contained_labware_bottom_first=[
            StackerStoredLabwareGroup(
                primaryLabwareId="labware-1",
                lidLabwareId="lid-1",
                adapterLabwareId=None,
            ),
            StackerStoredLabwareGroup(
                primaryLabwareId="labware-2",
                lidLabwareId="lid-2",
                adapterLabwareId=None,
            ),
            StackerStoredLabwareGroup(
                primaryLabwareId="labware-3",
                lidLabwareId="lid-3",
                adapterLabwareId=None,
            ),
        ],
        max_pool_count=5,
        pool_overlap=0,
        pool_height=0,
    )
    assert subject.get_contained_labware() == [
        StackerStoredLabwareGroup(
            primaryLabwareId="labware-1", lidLabwareId="lid-1", adapterLabwareId=None
        ),
        StackerStoredLabwareGroup(
            primaryLabwareId="labware-2", lidLabwareId="lid-2", adapterLabwareId=None
        ),
        StackerStoredLabwareGroup(
            primaryLabwareId="labware-3", lidLabwareId="lid-3", adapterLabwareId=None
        ),
    ]



"""Tests for delay properties in the Opentrons protocol API."""

from typing import Any, Union

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st
from pydantic import ValidationError

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    LiquidClassTouchTipParams,
    TouchTipProperties,
)

from . import (
    boolean_looking_values,
    invalid_values,
    negative_or_zero_floats_and_ints,
    positive_non_zero_floats_and_ints,
    reasonable_numbers,
)
from opentrons.protocol_api._liquid_properties import _build_touch_tip_properties


def test_touch_tip_properties_enable_and_disable() -> None:
    """Test enabling and disabling TouchTipProperties."""
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=1, mmFromEdge=1, speed=100),
        )
    )
    tp.enabled = True
    assert tp.enabled is True
    tp.enabled = False
    assert tp.enabled is False


def test_touch_tip_properties_none_instantiation_combos() -> None:
    """Test handling of None combinations in TouchTipProperties instantiation."""
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(
                enable=True,
                params=LiquidClassTouchTipParams(
                    zOffset=None,  # type: ignore
                    mmFromEdge=None,  # type: ignore
                    speed=None,  # type: ignore
                ),
            )
        )
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(
                enable=None,  # type: ignore
                params=LiquidClassTouchTipParams(zOffset=None, mmFromEdge=1, speed=1),  # type: ignore
            )
        )
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(
                enable=True,
                params=LiquidClassTouchTipParams(zOffset=1, mmFromEdge=None, speed=1),  # type: ignore
            )
        )
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(
                enable=True,
                params=LiquidClassTouchTipParams(zOffset=1, mmFromEdge=1, speed=None),  # type: ignore
            )
        )


@given(bad_value=st.one_of(invalid_values, boolean_looking_values))
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_enabled_bad_values(bad_value: Any) -> None:
    """Test bad values for TouchTipProperties.enabled."""
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(
                enable=bad_value,
                params=LiquidClassTouchTipParams(zOffset=1, mmFromEdge=1, speed=1),
            )
        )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=1, mmFromEdge=1, speed=1),
        )
    )
    with pytest.raises(ValueError):
        tp.enabled = bad_value


@given(good_value=reasonable_numbers)
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_z_offset(good_value: Union[int, float]) -> None:
    """Test valid z_offset."""
    _build_touch_tip_properties(
        TouchTipProperties(
            enable=True,
            params=LiquidClassTouchTipParams(
                zOffset=good_value, mmFromEdge=1, speed=10
            ),
        )
    )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=10),
        )
    )
    tp.z_offset = good_value
    assert tp.z_offset == float(good_value)


@given(bad_value=invalid_values)
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_z_offset_bad_values(bad_value: Any) -> None:
    """Test invalid z_offset values."""
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(
                enable=True,
                params=LiquidClassTouchTipParams(
                    zOffset=bad_value, mmFromEdge=1, speed=10
                ),
            )
        )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=10),
        )
    )
    with pytest.raises(ValueError):
        tp.z_offset = bad_value


@given(good_value=reasonable_numbers)
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_mm_from_edge(good_value: Union[int, float]) -> None:
    """Test valid mm_from_edge."""
    _build_touch_tip_properties(
        TouchTipProperties(
            enable=True,
            params=LiquidClassTouchTipParams(
                zOffset=0, mmFromEdge=good_value, speed=10
            ),
        )
    )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=10),
        )
    )
    tp.mm_from_edge = good_value
    assert tp.mm_from_edge == float(good_value)


@given(bad_value=invalid_values)
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_mm_from_edge_bad_values(bad_value: Any) -> None:
    """Test invalid mm_from_edge values."""
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(
                enable=True,
                params=LiquidClassTouchTipParams(
                    zOffset=bad_value, mmFromEdge=1, speed=10
                ),
            )
        )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=True,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=10),
        )
    )
    with pytest.raises(ValueError):
        tp.mm_from_edge = bad_value


@given(good_value=positive_non_zero_floats_and_ints)
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_speed(good_value: Union[int, float]) -> None:
    """Test valid speed."""
    _build_touch_tip_properties(
        TouchTipProperties(
            enable=True,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=good_value),
        )
    )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=10),
        )
    )
    tp.speed = good_value
    assert tp.speed == float(good_value)


@given(bad_value=st.one_of(invalid_values, negative_or_zero_floats_and_ints))
@settings(deadline=None, max_examples=50)
def test_touch_tip_properties_speed_bad_values(bad_value: Any) -> None:
    """Test invalid speed values."""
    with pytest.raises(ValidationError):
        _build_touch_tip_properties(
            TouchTipProperties(
                enable=True,
                params=LiquidClassTouchTipParams(
                    zOffset=0, mmFromEdge=1, speed=bad_value
                ),
            )
        )
    tp = _build_touch_tip_properties(
        TouchTipProperties(
            enable=False,
            params=LiquidClassTouchTipParams(zOffset=0, mmFromEdge=1, speed=10),
        )
    )
    with pytest.raises(ValueError):
        tp.speed = bad_value



"""Command models to open the latch of a Flex Stacker."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional

from pydantic import BaseModel, Field
from typing_extensions import Type

from ...errors import (
    ErrorOccurrence,
)
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData

if TYPE_CHECKING:
    from ...state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler

UnsafeFlexStackerOpenLatchCommandType = Literal["unsafe/flexStacker/openLatch"]


class UnsafeFlexStackerOpenLatchParams(BaseModel):
    """The parameters defining how a stacker should open its latch."""

    moduleId: str = Field(..., description="Unique ID of the Flex Stacker")


class UnsafeFlexStackerOpenLatchResult(BaseModel):
    """Result data from a stacker UnsafeFlexStackerOpenLatch command."""


class UnsafeFlexStackerOpenLatchImpl(
    AbstractCommandImpl[
        UnsafeFlexStackerOpenLatchParams, SuccessData[UnsafeFlexStackerOpenLatchResult]
    ]
):
    """Implementation of a stacker UnsafeFlexStackerOpenLatch command."""

    def __init__(
        self, state_view: StateView, equipment: EquipmentHandler, **kwargs: object
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(
        self, params: UnsafeFlexStackerOpenLatchParams
    ) -> SuccessData[UnsafeFlexStackerOpenLatchResult]:
        """Execute the stacker UnsafeFlexStackerOpenLatch command.

        Opening the latch could result in labware falling down the stacker hopper,
        resulting in an error state. This command should be used with care, outside of
        a protocol and followed by a close latch command.
        """
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )
        stacker_hw = self._equipment.get_module_hardware_api(stacker_state.module_id)

        if stacker_hw is not None:
            await stacker_hw.open_latch()
        return SuccessData(public=UnsafeFlexStackerOpenLatchResult())


class UnsafeFlexStackerOpenLatch(
    BaseCommand[
        UnsafeFlexStackerOpenLatchParams,
        UnsafeFlexStackerOpenLatchResult,
        ErrorOccurrence,
    ]
):
    """A command to UnsafeFlexStackerOpenLatch the Flex Stacker of labware."""

    commandType: UnsafeFlexStackerOpenLatchCommandType = "unsafe/flexStacker/openLatch"
    params: UnsafeFlexStackerOpenLatchParams
    result: Optional[UnsafeFlexStackerOpenLatchResult] = None

    _ImplementationCls: Type[UnsafeFlexStackerOpenLatchImpl] = (
        UnsafeFlexStackerOpenLatchImpl
    )


class UnsafeFlexStackerOpenLatchCreate(
    BaseCommandCreate[UnsafeFlexStackerOpenLatchParams]
):
    """A request to execute a Flex Stacker UnsafeFlexStackerOpenLatch command."""

    commandType: UnsafeFlexStackerOpenLatchCommandType = "unsafe/flexStacker/openLatch"
    params: UnsafeFlexStackerOpenLatchParams

    _CommandCls: Type[UnsafeFlexStackerOpenLatch] = UnsafeFlexStackerOpenLatch



"""Command models to capture an image with a camera."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, Tuple

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Literal, Type

from opentrons_shared_data.data_files import MimeType

from ..errors import (
    CameraDisabledError,
    CameraSettingsInvalidError,
    FileNameInvalidError,
)
from ..errors.error_occurrence import ErrorOccurrence
from ..resources import CameraProvider, FileProvider
from ..resources.camera_provider import ImageParameters
from ..resources.file_provider import (
    SPECIAL_CHARACTERS,
    ImageCaptureCmdFileNameMetadata,
)
from ..state import update_types
from ..types import PreconditionTypes
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from opentrons.system.camera import (
    BRIGHTNESS_DEFAULT,
    CONTRAST_DEFAULT,
    RESOLUTION_DEFAULT,
    RESOLUTION_MAX,
    RESOLUTION_MIN,
    SATURATION_DEFAULT,
    ZOOM_DEFAULT,
    ZOOM_MAX,
    ZOOM_MIN,
)

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


CaptureImageCommandType = Literal["captureImage"]


class CaptureImageParams(BaseModel):
    """Input parameters for an image capture."""

    fileName: str | SkipJsonSchema[None] = Field(
        None,
        description="Optional file name to use when storing the results of an Image Capture.",
        json_schema_extra=_remove_default,
    )
    resolution: Optional[Tuple[int, int]] = Field(
        None,
        description="Width by height resolution in pixels for the image to be captured with.",
    )
    zoom: Optional[float] = Field(
        None,
        description="Multiplier to use when cropping and scaling a captured image. Scale is 1.0 to 2.0.",
    )
    pan: Optional[Tuple[int, int]] = Field(
        None,
        description="X/Y (pixels) position to pan to for a given zoom. Default is the center of the image.",
    )
    contrast: Optional[float] = Field(
        None,
        description="The contrast to use when processing an image. Scale is 0% to 100%.",
    )
    brightness: Optional[float] = Field(
        None,
        description="The brightness to use when processing an image. Scale is 0% to 100%.",
    )
    saturation: Optional[float] = Field(
        None,
        description="The saturation to use when processing an image. Scale is 0% to 100%.",
    )


class CaptureImageResult(BaseModel):
    """Result data from running an image capture."""

    fileId: Optional[str] = Field(
        None,
        description="File ID for image files output as a result of an image capture action.",
    )
    resolution: Tuple[int, int] = Field(
        ...,
        description="Width by height resolution in pixels the image was captured with.",
    )
    zoom: float = Field(
        ...,
        description="Multiplier used when cropping and scaling the captured image. Scale is 1.0 to 2.0.",
    )
    pan: Optional[Tuple[int, int]] = Field(
        None,
        description="X/Y (pixels) position panned to.",
    )
    contrast: float = Field(
        ...,
        description="The contrast used when processing the image. Scale is 0% to 100%.",
    )
    brightness: float = Field(
        ...,
        description="The brightness used when processing the image. Scale is 0% to 100%.",
    )
    saturation: float = Field(
        ...,
        description="The saturation used when processing the image. Scale is 0% to 100%.",
    )


def _converted_image_params(params: CaptureImageParams) -> ImageParameters:
    return ImageParameters(
        resolution=params.resolution,
        zoom=params.zoom,
        pan=params.pan,
        contrast=(
            (params.contrast / 100) * 2.0 if params.contrast is not None else None
        ),
        brightness=(
            int(((params.brightness * 256) // 100) - 128) * -1
            if params.brightness is not None
            else None
        ),
        saturation=(
            (params.saturation / 100) * 2.0 if params.saturation is not None else None
        ),
    )


def _revert_image_parameters(
    file_id: Optional[str], image_params: ImageParameters
) -> CaptureImageResult:
    contrast = (
        image_params.contrast if image_params.contrast is not None else CONTRAST_DEFAULT
    )
    brightness = (
        image_params.brightness
        if image_params.brightness is not None
        else BRIGHTNESS_DEFAULT
    )
    saturation = (
        image_params.saturation
        if image_params.saturation is not None
        else SATURATION_DEFAULT
    )

    return CaptureImageResult(
        fileId=file_id,
        resolution=image_params.resolution
        if image_params.resolution is not None
        else RESOLUTION_DEFAULT,
        zoom=image_params.zoom if image_params.zoom is not None else ZOOM_DEFAULT,
        pan=image_params.pan,
        contrast=(contrast / 2) * 100.0,
        brightness=round((((brightness * -1) + 128) * 100) / 256),
        saturation=(saturation / 2) * 100.0,
    )


def _validate_image_params(params: CaptureImageParams) -> None:
    # Validate the filename param provided to fail analysis
    if params.fileName is not None and set(SPECIAL_CHARACTERS).intersection(
        set(params.fileName)
    ):
        raise FileNameInvalidError(
            message=f"Capture image filename cannot contain character(s): {SPECIAL_CHARACTERS.intersection(set(params.fileName))}"
        )

    # Validate the image filter parameters
    if params.zoom is not None and (params.zoom < ZOOM_MIN or params.zoom > ZOOM_MAX):
        raise CameraSettingsInvalidError(
            message="Capture image zoom must be a valid value from 1.0X to 2.0X zoom."
        )
    if params.resolution is not None and (
        params.resolution[0] < RESOLUTION_MIN[0]
        or params.resolution[1] < RESOLUTION_MIN[1]
        or params.resolution[0] > RESOLUTION_MAX[0]
        or params.resolution[1] > RESOLUTION_MAX[1]
    ):
        raise CameraSettingsInvalidError(
            message="Capture image resolution must be a valid resolution from 240p through 8K resolutuon."
        )
    if params.brightness is not None and (
        params.brightness < 0 or params.brightness > 100
    ):
        raise CameraSettingsInvalidError(
            message="Capture image brightness must be a percentage from 0% to 100%."
        )
    if params.contrast is not None and (params.contrast < 0 or params.contrast > 100):
        raise CameraSettingsInvalidError(
            message="Capture image contrast must be a percentage from 0% to 100%."
        )
    if params.saturation is not None and (
        params.saturation < 0 or params.saturation > 100
    ):
        raise CameraSettingsInvalidError(
            message="Capture image saturation must be a percentage from 0% to 100%."
        )


class CaptureImageImpl(
    AbstractCommandImpl[CaptureImageParams, SuccessData[CaptureImageResult]]
):
    """Execution implementation of an image capture."""

    def __init__(
        self,
        state_view: StateView,
        file_provider: FileProvider,
        camera_provider: CameraProvider,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._file_provider = file_provider
        self._camera_provider = camera_provider

    async def execute(
        self, params: CaptureImageParams
    ) -> SuccessData[CaptureImageResult]:
        """Initiate an image capture with a camera."""
        state_update = update_types.StateUpdate()
        state_update.precondition_update = update_types.PreconditionUpdate(
            {PreconditionTypes.IS_CAMERA_USED: True}
        )

        # Validate that the provided parameters are all acceptable. We do this here and in system/camera.py to ensure analysis fails properly.
        _validate_image_params(params)

        # Handle capturing an image with the CameraProvider - Engine camera settings take priority
        camera_settings = await self._camera_provider.get_camera_settings()
        engine_camera_settings = self._state_view.camera.get_enablement_settings()
        if (
            engine_camera_settings is None and camera_settings.cameraEnabled is False
        ) or (
            engine_camera_settings is not None
            and engine_camera_settings.cameraEnabled is False
        ):
            raise CameraDisabledError(
                "Cannot capture image because Camera is disabled."
            )

        parameters = _converted_image_params(params=params)
        camera_data = await self._camera_provider.capture_image(
            self._state_view.config.robot_type, parameters
        )

        # Conditionally save file if camera data was returned - in simulation we don't return anything.
        file_id: str | None = None
        if camera_data:
            this_cmd_id = self._state_view.commands.get_running_command_id()
            prev_cmd = self._state_view.commands.get_most_recently_finalized_command()
            prev_cmd_id = prev_cmd.command.id if prev_cmd is not None else None

            file_info = await self._file_provider.write_file(
                data=camera_data,
                mime_type=MimeType.IMAGE_JPEG,
                command_metadata=ImageCaptureCmdFileNameMetadata(
                    step_number=len(self._state_view.commands.get_all()),
                    command_timestamp=datetime.now(),
                    base_filename=params.fileName,
                    command_id=this_cmd_id or "",
                    prev_command_id=prev_cmd_id or "",
                ),
            )
            file_id = file_info.id
            state_update.files_added = update_types.FilesAddedUpdate(file_ids=[file_id])

        result = _revert_image_parameters(file_id=file_id, image_params=parameters)

        return SuccessData(
            public=result,
            state_update=state_update,
        )


class CaptureImage(
    BaseCommand[CaptureImageParams, CaptureImageResult, ErrorOccurrence]
):
    """A command to execute an Absorbance Reader measurement."""

    commandType: CaptureImageCommandType = "captureImage"
    params: CaptureImageParams
    result: Optional[CaptureImageResult] = None

    _ImplementationCls: Type[CaptureImageImpl] = CaptureImageImpl


class CaptureImageCreate(BaseCommandCreate[CaptureImageParams]):
    """A request to execute an Absorbance Reader measurement."""

    commandType: CaptureImageCommandType = "captureImage"
    params: CaptureImageParams

    _CommandCls: Type[CaptureImage] = CaptureImage



from typing import Dict, Iterator, List, Tuple, Union, cast

import pytest

from opentrons_shared_data.pipette.load_data import load_definition
from opentrons_shared_data.pipette.pipette_definition import (
    PipetteConfigurations,
    ValidNozzleMaps,
)
from opentrons_shared_data.pipette.types import (
    PipetteChannelType,
    PipetteModelType,
    PipetteOEMType,
    PipetteVersionType,
)
from tests.opentrons.protocol_engine.pipette_fixtures import (
    EIGHT_CHANNEL_COLS,
    NINETY_SIX_COLS,
    NINETY_SIX_ROWS,
)

from opentrons.hardware_control import nozzle_manager
from opentrons.types import NozzleConfigurationType, Point

# Ninety six channel valid nozzle maps
NINETY_SIX_FULL: Dict[str, List[str]] = {
    "Full": sum(
        [
            NINETY_SIX_ROWS["A"],
            NINETY_SIX_ROWS["B"],
            NINETY_SIX_ROWS["C"],
            NINETY_SIX_ROWS["D"],
            NINETY_SIX_ROWS["E"],
            NINETY_SIX_ROWS["F"],
            NINETY_SIX_ROWS["G"],
            NINETY_SIX_ROWS["H"],
        ],
        [],
    )
}
NINETY_SIX_COL_1: Dict[str, List[str]] = {"Column1": NINETY_SIX_COLS["1"]}
NINETY_SIX_COL_12: Dict[str, List[str]] = {"Column12": NINETY_SIX_COLS["12"]}
NINETY_SIX_ROW_A: Dict[str, List[str]] = {"RowA": NINETY_SIX_ROWS["A"]}
NINETY_SIX_ROW_H: Dict[str, List[str]] = {"RowH": NINETY_SIX_ROWS["H"]}

A1_D6: Dict[str, List[str]] = {
    "A1_D6": [
        "A1",
        "A2",
        "A3",
        "A4",
        "A5",
        "A6",
        "B1",
        "B2",
        "B3",
        "B4",
        "B5",
        "B6",
        "C1",
        "C2",
        "C3",
        "C4",
        "C5",
        "C6",
        "D1",
        "D2",
        "D3",
        "D4",
        "D5",
        "D6",
    ]
}
E7_H12: Dict[str, List[str]] = {
    "E7_H12": [
        "E7",
        "E8",
        "E9",
        "E10",
        "E11",
        "E12",
        "F7",
        "F8",
        "F9",
        "F10",
        "F11",
        "F12",
        "G7",
        "G8",
        "G9",
        "G10",
        "G11",
        "G12",
        "H7",
        "H8",
        "H9",
        "H10",
        "H11",
        "H12",
    ]
}
E1_H6: Dict[str, List[str]] = {
    "E1_H6": [
        "E1",
        "E2",
        "E3",
        "E4",
        "E5",
        "E6",
        "F1",
        "F2",
        "F3",
        "F4",
        "F5",
        "F6",
        "G1",
        "G2",
        "G3",
        "G4",
        "G5",
        "G6",
        "H1",
        "H2",
        "H3",
        "H4",
        "H5",
        "H6",
    ]
}
A1_B12: Dict[str, List[str]] = {
    "A1_B12": [
        "A1",
        "A2",
        "A3",
        "A4",
        "A5",
        "A6",
        "A7",
        "A8",
        "A9",
        "A10",
        "A11",
        "A12",
        "B1",
        "B2",
        "B3",
        "B4",
        "B5",
        "B6",
        "B7",
        "B8",
        "B9",
        "B10",
        "B11",
        "B12",
    ]
}
G1_H12: Dict[str, List[str]] = {
    "G1_H12": [
        "G1",
        "G2",
        "G3",
        "G4",
        "G5",
        "G6",
        "G7",
        "G8",
        "G9",
        "G10",
        "G11",
        "G12",
        "H1",
        "H2",
        "H3",
        "H4",
        "H5",
        "H6",
        "H7",
        "H8",
        "H9",
        "H10",
        "H11",
        "H12",
    ]
}
A1_H3: Dict[str, List[str]] = {
    "A1_H3": [
        "A1",
        "A2",
        "A3",
        "B1",
        "B2",
        "B3",
        "C1",
        "C2",
        "C3",
        "D1",
        "D2",
        "D3",
        "E1",
        "E2",
        "E3",
        "F1",
        "F2",
        "F3",
        "G1",
        "G2",
        "G3",
        "H1",
        "H2",
        "H3",
    ]
}
A10_H12: Dict[str, List[str]] = {
    "A10_H12": [
        "A10",
        "A11",
        "A12",
        "B10",
        "B11",
        "B12",
        "C10",
        "C11",
        "C12",
        "D10",
        "D11",
        "D12",
        "E10",
        "E11",
        "E12",
        "F10",
        "F11",
        "F12",
        "G10",
        "G11",
        "G12",
        "H10",
        "H11",
        "H12",
    ]
}

# Eight channel valid nozzle maps
EIGHT_CHANNEL_FULL: Dict[str, List[str]] = {"Full": EIGHT_CHANNEL_COLS["1"]}
A1_D1: Dict[str, List[str]] = {"A1_D1": ["A1", "B1", "C1", "D1"]}
E1_H1: Dict[str, List[str]] = {"E1_H1": ["E1", "F1", "G1", "H1"]}
A1: Dict[str, List[str]] = {"A1": ["A1"]}
H1: Dict[str, List[str]] = {"H1": ["H1"]}


@pytest.mark.parametrize(
    "pipette_details",
    [
        (PipetteModelType.p10, PipetteVersionType(major=1, minor=3)),
        (PipetteModelType.p20, PipetteVersionType(major=2, minor=0)),
        (PipetteModelType.p50, PipetteVersionType(major=3, minor=4)),
        (PipetteModelType.p300, PipetteVersionType(major=2, minor=1)),
        (PipetteModelType.p1000, PipetteVersionType(major=3, minor=5)),
    ],
)
def test_single_pipettes_always_full(
    pipette_details: Tuple[PipetteModelType, PipetteVersionType],
) -> None:
    config = load_definition(
        pipette_details[0],
        PipetteChannelType.SINGLE_CHANNEL,
        pipette_details[1],
        PipetteOEMType.OT,
    )
    subject = nozzle_manager.NozzleConfigurationManager.build_from_config(
        config, ValidNozzleMaps(maps=A1)
    )
    assert subject.current_configuration.configuration == NozzleConfigurationType.FULL

    subject.update_nozzle_configuration("A1", "A1", "A1")
    assert subject.current_configuration.configuration == NozzleConfigurationType.FULL

    subject.reset_to_default_configuration()
    assert subject.current_configuration.configuration == NozzleConfigurationType.FULL


@pytest.mark.parametrize(
    "pipette_details",
    [
        (PipetteModelType.p10, PipetteVersionType(major=1, minor=3)),
        (PipetteModelType.p20, PipetteVersionType(major=2, minor=0)),
        (PipetteModelType.p50, PipetteVersionType(major=3, minor=4)),
        (PipetteModelType.p300, PipetteVersionType(major=2, minor=1)),
        (PipetteModelType.p1000, PipetteVersionType(major=3, minor=5)),
    ],
)
def test_single_pipette_map_entries(
    pipette_details: Tuple[PipetteModelType, PipetteVersionType],
) -> None:
    config = load_definition(
        pipette_details[0],
        PipetteChannelType.SINGLE_CHANNEL,
        pipette_details[1],
        PipetteOEMType.OT,
    )
    subject = nozzle_manager.NozzleConfigurationManager.build_from_config(
        config, ValidNozzleMaps(maps=A1)
    )

    def test_map_entries(nozzlemap: nozzle_manager.NozzleMap) -> None:
        assert nozzlemap.back_left == "A1"
        assert nozzlemap.front_right == "A1"
        assert list(nozzlemap.map_store.keys()) == ["A1"]
        assert list(nozzlemap.rows.keys()) == ["A"]
        assert list(nozzlemap.columns.keys()) == ["1"]
        assert nozzlemap.rows["A"] == ["A1"]
        assert nozzlemap.columns["1"] == ["A1"]
        assert nozzlemap.tip_count == 1

    test_map_entries(subject.current_configuration)
    subject.update_nozzle_configuration("A1", "A1", "A1")
    test_map_entries(subject.current_configuration)
    subject.reset_to_default_configuration()
    test_map_entries(subject.current_configuration)


@pytest.mark.parametrize(
    "pipette_details",
    [
        (PipetteModelType.p10, PipetteVersionType(major=1, minor=3)),
        (PipetteModelType.p20, PipetteVersionType(major=2, minor=0)),
        (PipetteModelType.p50, PipetteVersionType(major=3, minor=4)),
        (PipetteModelType.p300, PipetteVersionType(major=2, minor=1)),
        (PipetteModelType.p1000, PipetteVersionType(major=3, minor=5)),
    ],
)
def test_single_pipette_map_geometry(
    pipette_details: Tuple[PipetteModelType, PipetteVersionType],
) -> None:
    config = load_definition(
        pipette_details[0],
        PipetteChannelType.SINGLE_CHANNEL,
        pipette_details[1],
        PipetteOEMType.OT,
    )
    subject = nozzle_manager.NozzleConfigurationManager.build_from_config(
        config, ValidNozzleMaps(maps=A1)
    )

    def test_map_geometry(nozzlemap: nozzle_manager.NozzleMap) -> None:
        assert nozzlemap.xy_center_offset == Point(*config.nozzle_map["A1"])
        assert nozzlemap.y_center_offset == Point(*config.nozzle_map["A1"])
        assert nozzlemap.front_nozzle_offset == Point(*config.nozzle_map["A1"])
        assert nozzlemap.starting_nozzle_offset == Point(*config.nozzle_map["A1"])

    test_map_geometry(subject.current_configuration)
    subject.update_nozzle_configuration("A1", "A1", "A1")
    test_map_geometry(subject.current_configuration)
    subject.reset_to_default_configuration()
    test_map_geometry(subject.current_configuration)


@pytest.mark.parametrize(
    "pipette_details",
    [
        (PipetteModelType.p10, PipetteVersionType(major=1, minor=3)),
        (PipetteModelType.p20, PipetteVersionType(major=2, minor=0)),
        (PipetteModelType.p50, PipetteVersionType(major=3, minor=4)),
        (PipetteModelType.p300, PipetteVersionType(major=2, minor=1)),
        (PipetteModelType.p1000, PipetteVersionType(major=3, minor=5)),
    ],
)
def test_multi_config_identification(
    pipette_details: Tuple[PipetteModelType, PipetteVersionType],
) -> None:
    config = load_definition(
        pipette_details[0],
        PipetteChannelType.EIGHT_CHANNEL,
        pipette_details[1],
        PipetteOEMType.OT,
    )
    subject = nozzle_manager.NozzleConfigurationManager.build_from_config(
        config,
        ValidNozzleMaps(maps=EIGHT_CHANNEL_FULL | A1_D1 | A1 | H1),
    )

    assert subject.current_configuration.configuration == NozzleConfigurationType.FULL

    subject.update_nozzle_configuration("A1", "H1", "A1")
    assert subject.current_configuration.configuration == NozzleConfigurationType.FULL

    subject.reset_to_default_configuration()
    assert subject.current_configuration.configuration == NozzleConfigurationType.FULL

    subject.update_nozzle_configuration("A1", "D1", "A1")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.COLUMN
    )

    subject.update_nozzle_configuration("A1", "A1", "A1")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.SINGLE
    )

    subject.update_nozzle_configuration("H1", "H1", "H1")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.SINGLE
    )

    subject.reset_to_default_configuration()
    assert subject.current_configuration.configuration == NozzleConfigurationType.FULL


@pytest.mark.parametrize(
    "pipette_details",
    [
        (PipetteModelType.p10, PipetteVersionType(major=1, minor=3)),
        (PipetteModelType.p20, PipetteVersionType(major=2, minor=0)),
        (PipetteModelType.p50, PipetteVersionType(major=3, minor=4)),
        (PipetteModelType.p300, PipetteVersionType(major=2, minor=1)),
        (PipetteModelType.p1000, PipetteVersionType(major=3, minor=5)),
    ],
)
def test_multi_config_map_entries(
    pipette_details: Tuple[PipetteModelType, PipetteVersionType],
) -> None:
    config = load_definition(
        pipette_details[0],
        PipetteChannelType.EIGHT_CHANNEL,
        pipette_details[1],
        PipetteOEMType.OT,
    )
    subject = nozzle_manager.NozzleConfigurationManager.build_from_config(
        config,
        ValidNozzleMaps(maps=EIGHT_CHANNEL_FULL | A1_D1 | A1 | H1),
    )

    def test_map_entries(
        nozzlemap: nozzle_manager.NozzleMap, nozzles: List[str]
    ) -> None:
        assert nozzlemap.back_left == nozzles[0]
        assert nozzlemap.front_right == nozzles[-1]
        assert list(nozzlemap.map_store.keys()) == nozzles
        assert list(nozzlemap.rows.keys()) == [nozzle[0] for nozzle in nozzles]
        assert list(nozzlemap.columns.keys()) == ["1"]
        for rowname, row_elements in nozzlemap.rows.items():
            assert row_elements == [f"{rowname}1"]

        assert nozzlemap.columns["1"] == nozzles
        assert nozzlemap.tip_count == len(nozzles)

    test_map_entries(
        subject.current_configuration, ["A1", "B1", "C1", "D1", "E1", "F1", "G1", "H1"]
    )
    subject.update_nozzle_configuration("A1", "H1", "A1")
    test_map_entries(
        subject.current_configuration, ["A1", "B1", "C1", "D1", "E1", "F1", "G1", "H1"]
    )
    subject.update_nozzle_configuration("A1", "D1", "A1")
    test_map_entries(subject.current_configuration, ["A1", "B1", "C1", "D1"])
    subject.update_nozzle_configuration("A1", "A1", "A1")
    test_map_entries(subject.current_configuration, ["A1"])
    subject.update_nozzle_configuration("H1", "H1", "H1")
    test_map_entries(subject.current_configuration, ["H1"])


def assert_offset_in_center_of(
    offset: Point, between: Union[Tuple[str, str], str], config: PipetteConfigurations
) -> None:
    if isinstance(between, str):
        assert offset == Point(*config.nozzle_map[between])
    else:
        assert (
            offset
            == (
                Point(*config.nozzle_map[between[0]])
                + Point(*config.nozzle_map[between[1]])
            )
            * 0.5
        )


@pytest.mark.parametrize(
    "pipette_details",
    [
        (PipetteModelType.p10, PipetteVersionType(major=1, minor=3)),
        (PipetteModelType.p20, PipetteVersionType(major=2, minor=0)),
        (PipetteModelType.p50, PipetteVersionType(major=3, minor=4)),
        (PipetteModelType.p300, PipetteVersionType(major=2, minor=1)),
        (PipetteModelType.p1000, PipetteVersionType(major=3, minor=5)),
    ],
)
def test_multi_config_geometry(
    pipette_details: Tuple[PipetteModelType, PipetteVersionType],
) -> None:
    config = load_definition(
        pipette_details[0],
        PipetteChannelType.EIGHT_CHANNEL,
        pipette_details[1],
        PipetteOEMType.OT,
    )
    subject = nozzle_manager.NozzleConfigurationManager.build_from_config(
        config,
        ValidNozzleMaps(maps=EIGHT_CHANNEL_FULL | A1_D1 | E1_H1 | A1 | H1),
    )

    def test_map_geometry(
        nozzlemap: nozzle_manager.NozzleMap,
        front_nozzle: str,
        starting_nozzle: str,
        xy_center_in_center_of: Union[Tuple[str, str], str],
        y_center_in_center_of: Union[Tuple[str, str], str],
    ) -> None:
        assert_offset_in_center_of(
            nozzlemap.xy_center_offset, xy_center_in_center_of, config
        )
        assert_offset_in_center_of(
            nozzlemap.y_center_offset, y_center_in_center_of, config
        )

        assert nozzlemap.front_nozzle_offset == Point(*config.nozzle_map[front_nozzle])
        assert nozzlemap.starting_nozzle_offset == Point(
            *config.nozzle_map[starting_nozzle]
        )

    test_map_geometry(
        subject.current_configuration, "H1", "A1", ("A1", "H1"), ("A1", "H1")
    )

    subject.update_nozzle_configuration("A1", "A1", "A1")
    test_map_geometry(subject.current_configuration, "A1", "A1", "A1", "A1")

    subject.update_nozzle_configuration("E1", "H1", "E1")
    test_map_geometry(
        subject.current_configuration, "H1", "E1", ("E1", "H1"), ("E1", "H1")
    )

    subject.reset_to_default_configuration()
    test_map_geometry(
        subject.current_configuration, "H1", "A1", ("A1", "H1"), ("A1", "H1")
    )


@pytest.mark.parametrize(
    "pipette_details", [(PipetteModelType.p1000, PipetteVersionType(major=3, minor=5))]
)
def test_96_config_identification(
    pipette_details: Tuple[PipetteModelType, PipetteVersionType],
) -> None:
    config = load_definition(
        pipette_details[0],
        PipetteChannelType.NINETY_SIX_CHANNEL,
        pipette_details[1],
        PipetteOEMType.OT,
    )
    subject = nozzle_manager.NozzleConfigurationManager.build_from_config(
        config,
        ValidNozzleMaps(
            maps=NINETY_SIX_FULL
            | NINETY_SIX_COL_1
            | NINETY_SIX_COL_12
            | NINETY_SIX_ROW_A
            | NINETY_SIX_ROW_H
            | A1_D6
            | E7_H12
            | E1_H6
            | A1_B12
            | G1_H12
            | A1_H3
            | A10_H12
        ),
    )

    assert subject.current_configuration.configuration == NozzleConfigurationType.FULL
    subject.update_nozzle_configuration("A1", "H12")
    assert subject.current_configuration.configuration == NozzleConfigurationType.FULL
    subject.update_nozzle_configuration("A1", "H1")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.COLUMN
    )
    subject.update_nozzle_configuration("A12", "H12")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.COLUMN
    )

    subject.update_nozzle_configuration("A1", "A12")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.ROW
    )
    subject.update_nozzle_configuration("H1", "H12")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.ROW
    )

    subject.update_nozzle_configuration("E1", "H6")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.SUBRECT
    )
    subject.update_nozzle_configuration("E7", "H12")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.SUBRECT
    )

    subject.update_nozzle_configuration("A1", "B12")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.SUBRECT
    )
    subject.update_nozzle_configuration("G1", "H12")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.SUBRECT
    )
    subject.update_nozzle_configuration("A1", "H3")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.SUBRECT
    )
    subject.update_nozzle_configuration("A10", "H12")
    assert (
        cast(
            NozzleConfigurationType,
            subject.current_configuration.configuration,
        )
        == NozzleConfigurationType.SUBRECT
    )


@pytest.mark.parametrize(
    "pipette_details", [(PipetteModelType.p1000, PipetteVersionType(major=3, minor=5))]
)
def test_96_config_map_entries(
    pipette_details: Tuple[PipetteModelType, PipetteVersionType],
) -> None:
    config = load_definition(
        pipette_details[0],
        PipetteChannelType.NINETY_SIX_CHANNEL,
        pipette_details[1],
        PipetteOEMType.OT,
    )
    subject = nozzle_manager.NozzleConfigurationManager.build_from_config(
        config,
        ValidNozzleMaps(
            maps=NINETY_SIX_FULL
            | NINETY_SIX_COL_1
            | NINETY_SIX_COL_12
            | NINETY_SIX_ROW_A
            | NINETY_SIX_ROW_H
            | A1_D6
            | E7_H12
            | E1_H6
            | A1_B12
            | G1_H12
            | A1_H3
            | A10_H12
        ),
    )

    def test_map_entries(
        nozzlemap: nozzle_manager.NozzleMap,
        rows: Dict[str, List[str]],
        cols: Dict[str, List[str]],
    ) -> None:
        assert nozzlemap.back_left == next(iter(rows.values()))[0]
        assert nozzlemap.front_right == next(reversed(list(rows.values())))[-1]

        def _nozzles() -> Iterator[str]:
            for row in rows.values():
                for nozzle in row:
                    yield nozzle

        assert list(nozzlemap.map_store.keys()) == list(_nozzles())
        assert nozzlemap.rows == rows
        assert nozzlemap.columns == cols
        assert nozzlemap.tip_count == sum(len(row) for row in rows.values())

    test_map_entries(
        subject.current_configuration,
        {
            "A": [
                "A1",
                "A2",
                "A3",
                "A4",
                "A5",
                "A6",
                "A7",
                "A8",
                "A9",
                "A10",
                "A11",
                "A12",
            ],
            "B": [
                "B1",
                "B2",
                "B3",
                "B4",
                "B5",
                "B6",
                "B7",
                "B8",
                "B9",
                "B10",
                "B11",
                "B12",
            ],
            "C": [
                "C1",
                "C2",
                "C3",
                "C4",
                "C5",
                "C6",
                "C7",
                "C8",
                "C9",
                "C10",
                "C11",
                "C12",
            ],
            "D": [
                "D1",
                "D2",
                "D3",
                "D4",
                "D5",
                "D6",
                "D7",
                "D8",
                "D9",
                "D10",
                "D11",
                "D12",
            ],
            "E": [
                "E1",
                "E2",
                "E3",
                "E4",
                "E5",
                "E6",
                "E7",
                "E8",
                "E9",
                "E10",
                "E11",
                "E12",
            ],
            "F": [
                "F1",
                "F2",
                "F3",
                "F4",
                "F5",
                "F6",
                "F7",
                "F8",
                "F9",
                "F10",
                "F11",
                "F12",
            ],
            "G": [
                "G1",
                "G2",
                "G3",
                "G4",
                "G5",
                "G6",
                "G7",
                "G8",
                "G9",
                "G10",
                "G11",
                "G12",
            ],
            "H": [
                "H1",
                "H2",
                "H3",
                "H4",
                "H5",
                "H6",
                "H7",
                "H8",
                "H9",
                "H10",
                "H11",
                "H12",
            ],
        },
        {
            "1": ["A1", "B1", "C1", "D1", "E1", "F1", "G1", "H1"],
            "2": ["A2", "B2", "C2", "D2", "E2", "F2", "G2", "H2"],
            "3": ["A3", "B3", "C3", "D3", "E3", "F3", "G3", "H3"],
            "4": ["A4", "B4", "C4", "D4", "E4", "F4", "G4", "H4"],
            "5": ["A5", "B5", "C5", "D5", "E5", "F5", "G5", "H5"],
            "6": ["A6", "B6", "C6", "D6", "E6", "F6", "G6", "H6"],
            "7": ["A7", "B7", "C7", "D7", "E7", "F7", "G7", "H7"],
            "8": ["A8", "B8", "C8", "D8", "E8", "F8", "G8", "H8"],
            "9": ["A9", "B9", "C9", "D9", "E9", "F9", "G9", "H9"],
            "10": ["A10", "B10", "C10", "D10", "E10", "F10", "G10", "H10"],
            "11": ["A11", "B11", "C11", "D11", "E11", "F11", "G11", "H11"],
            "12": ["A12", "B12", "C12", "D12", "E12", "F12", "G12", "H12"],
        },
    )

    subject.update_nozzle_configuration("A1", "H1")
    test_map_entries(
        subject.current_configuration,
        {
            "A": ["A1"],
            "B": ["B1"],
            "C": ["C1"],
            "D": ["D1"],
            "E": ["E1"],
            "F": ["F1"],
            "G": ["G1"],
            "H": ["H1"],
        },
        {"1": ["A1", "B1", "C1", "D1", "E1", "F1", "G1", "H1"]},
    )

    subject.update_nozzle_configuration("A12", "H12")
    test_map_entries(
        subject.current_configuration,
        {
            "A": ["A12"],
            "B": ["B12"],
            "C": ["C12"],
            "D": ["D12"],
            "E": ["E12"],
            "F": ["F12"],
            "G": ["G12"],
            "H": ["H12"],
        },
        {"12": ["A12", "B12", "C12", "D12", "E12", "F12", "G12", "H12"]},
    )

    subject.update_nozzle_configuration("A1", "A12")
    test_map_entries(
        subject.current_configuration,
        {
            "A": [
                "A1",
                "A2",
                "A3",
                "A4",
                "A5",
                "A6",
                "A7",
                "A8",
                "A9",
                "A10",
                "A11",
                "A12",
            ]
        },
        {
            "1": ["A1"],
            "2": ["A2"],
            "3": ["A3"],
            "4": ["A4"],
            "5": ["A5"],
            "6": ["A6"],
            "7": ["A7"],
            "8": ["A8"],
            "9": ["A9"],
            "10": ["A10"],
            "11": ["A11"],
            "12": ["A12"],
        },
    )

    subject.update_nozzle_configuration("H1", "H12")
    test_map_entries(
        subject.current_configuration,
        {
            "H": [
                "H1",
                "H2",
                "H3",
                "H4",
                "H5",
                "H6",
                "H7",
                "H8",
                "H9",
                "H10",
                "H11",
                "H12",
            ]
        },
        {
            "1": ["H1"],
            "2": ["H2"],
            "3": ["H3"],
            "4": ["H4"],
            "5": ["H5"],
            "6": ["H6"],
            "7": ["H7"],
            "8": ["H8"],
            "9": ["H9"],
            "10": ["H10"],
            "11": ["H11"],
            "12": ["H12"],
        },
    )

    subject.update_nozzle_configuration("A1", "D6")
    test_map_entries(
        subject.current_configuration,
        {
            "A": ["A1", "A2", "A3", "A4", "A5", "A6"],
            "B": ["B1", "B2", "B3", "B4", "B5", "B6"],
            "C": ["C1", "C2", "C3", "C4", "C5", "C6"],
            "D": ["D1", "D2", "D3", "D4", "D5", "D6"],
        },
        {
            "1": ["A1", "B1", "C1", "D1"],
            "2": ["A2", "B2", "C2", "D2"],
            "3": ["A3", "B3", "C3", "D3"],
            "4": ["A4", "B4", "C4", "D4"],
            "5": ["A5", "B5", "C5", "D5"],
            "6": ["A6", "B6", "C6", "D6"],
        },
    )

    subject.update_nozzle_configuration("E1", "H6")
    test_map_entries(
        subject.current_configuration,
        {
            "E": ["E1", "E2", "E3", "E4", "E5", "E6"],
            "F": ["F1", "F2", "F3", "F4", "F5", "F6"],
            "G": ["G1", "G2", "G3", "G4", "G5", "G6"],
            "H": ["H1", "H2", "H3", "H4", "H5", "H6"],
        },
        {
            "1": ["E1", "F1", "G1", "H1"],
            "2": ["E2", "F2", "G2", "H2"],
            "3": ["E3", "F3", "G3", "H3"],
            "4": ["E4", "F4", "G4", "H4"],
            "5": ["E5", "F5", "G5", "H5"],
            "6": ["E6", "F6", "G6", "H6"],
        },
    )

    subject.update_nozzle_configuration("E7", "H12")
    test_map_entries(
        subject.current_configuration,
        {
            "E": ["E7", "E8", "E9", "E10", "E11", "E12"],
            "F": ["F7", "F8", "F9", "F10", "F11", "F12"],
            "G": ["G7", "G8", "G9", "G10", "G11", "G12"],
            "H": ["H7", "H8", "H9", "H10", "H11", "H12"],
        },
        {
            "7": ["E7", "F7", "G7", "H7"],
            "8": ["E8", "F8", "G8", "H8"],
            "9": ["E9", "F9", "G9", "H9"],
            "10": ["E10", "F10", "G10", "H10"],
            "11": ["E11", "F11", "G11", "H11"],
            "12": ["E12", "F12", "G12", "H12"],
        },
    )


@pytest.mark.parametrize(
    "pipette_details", [(PipetteModelType.p1000, PipetteVersionType(major=3, minor=5))]
)
def test_96_config_geometry(
    pipette_details: Tuple[PipetteModelType, PipetteVersionType],
) -> None:
    config = load_definition(
        pipette_details[0],
        PipetteChannelType.NINETY_SIX_CHANNEL,
        pipette_details[1],
        PipetteOEMType.OT,
    )
    subject = nozzle_manager.NozzleConfigurationManager.build_from_config(
        config,
        ValidNozzleMaps(
            maps=NINETY_SIX_FULL
            | NINETY_SIX_COL_1
            | NINETY_SIX_COL_12
            | NINETY_SIX_ROW_A
            | NINETY_SIX_ROW_H
            | A1_D6
            | E7_H12
            | E1_H6
            | A1_B12
            | G1_H12
            | A1_H3
            | A10_H12
        ),
    )

    def test_map_geometry(
        config: PipetteConfigurations,
        nozzlemap: nozzle_manager.NozzleMap,
        starting_nozzle: str,
        front_nozzle: str,
        xy_center_between: Union[str, Tuple[str, str]],
        y_center_between: Union[str, Tuple[str, str]],
    ) -> None:
        assert_offset_in_center_of(
            nozzlemap.xy_center_offset, xy_center_between, config
        )
        assert_offset_in_center_of(nozzlemap.y_center_offset, y_center_between, config)

        assert nozzlemap.front_nozzle_offset == Point(*config.nozzle_map[front_nozzle])
        assert nozzlemap.starting_nozzle_offset == Point(
            *config.nozzle_map[starting_nozzle]
        )

    test_map_geometry(
        config, subject.current_configuration, "A1", "H1", ("A1", "H12"), ("A1", "H1")
    )

    subject.update_nozzle_configuration("A1", "H1")
    test_map_geometry(
        config, subject.current_configuration, "A1", "H1", ("A1", "H1"), ("A1", "H1")
    )

    subject.update_nozzle_configuration("A12", "H12")
    test_map_geometry(
        config,
        subject.current_configuration,
        "A12",
        "H12",
        ("A12", "H12"),
        ("A12", "H12"),
    )

    subject.update_nozzle_configuration("A1", "A12")
    test_map_geometry(
        config, subject.current_configuration, "A1", "A1", ("A1", "A12"), "A1"
    )

    subject.update_nozzle_configuration("H1", "H12")
    test_map_geometry(
        config, subject.current_configuration, "H1", "H1", ("H1", "H12"), "H1"
    )

    subject.update_nozzle_configuration("A1", "D6")
    test_map_geometry(
        config, subject.current_configuration, "A1", "D1", ("A1", "D6"), ("A1", "D1")
    )

    subject.update_nozzle_configuration("E7", "H12")
    test_map_geometry(
        config, subject.current_configuration, "E7", "H7", ("E7", "H12"), ("E7", "H7")
    )



"""Tests for opentrons.protocol_api.core.engine.LabwareCore."""

from typing import cast

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition2,
    LabwareRole,
    RectangularWellDefinition2,
)
from opentrons_shared_data.labware.labware_definition import (
    Metadata as LabwareDefinitionMetadata,
)
from opentrons_shared_data.labware.labware_definition import (
    Parameters2 as LabwareDefinition2Parameters,
)
from opentrons_shared_data.labware.types import (
    LabwareDefinition as LabwareDefDict,
)
from opentrons_shared_data.labware.types import (
    LabwareUri,
)

from opentrons.calibration_storage.helpers import uri_from_details
from opentrons.protocol_api._liquid import Liquid
from opentrons.protocol_api.core.engine import LabwareCore, WellCore
from opentrons.protocol_api.core.labware import LabwareLoadParams
from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.clients import SyncClient as EngineClient
from opentrons.protocol_engine.errors import LabwareNotOnDeckError
from opentrons.protocol_engine.types import (
    LabwareOffsetCreate,
    LabwareOffsetLocationSequence,
    LabwareOffsetVector,
    OnAddressableAreaOffsetLocationSequenceComponent,
    TipRackWellState,
)
from opentrons.types import DeckSlotName, Point


@pytest.fixture
def labware_definition() -> LabwareDefinition2:
    """Get a LabwareDefinition value object to use in tests."""
    return LabwareDefinition2.model_construct(ordering=[])  # type: ignore[call-arg]


@pytest.fixture
def mock_engine_client(
    decoy: Decoy, labware_definition: LabwareDefinition2
) -> EngineClient:
    """Get a mock ProtocolEngine synchronous client."""
    engine_client = decoy.mock(cls=EngineClient)
    decoy.when(engine_client.state.labware.get_definition("cool-labware")).then_return(
        labware_definition
    )
    return engine_client


@pytest.fixture
def subject(mock_engine_client: EngineClient) -> LabwareCore:
    """Get a LabwareCore test subject with mocked out dependencies."""
    return LabwareCore(labware_id="cool-labware", engine_client=mock_engine_client)


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            namespace="hello",
            version=42,
            parameters=LabwareDefinition2Parameters.model_construct(loadName="world"),  # type: ignore[call-arg]
            ordering=[],
        )
    ],
)
def test_get_load_params(subject: LabwareCore) -> None:
    """It should be able to get the definition's load parameters."""
    assert subject.get_load_params() == LabwareLoadParams("hello", "world", 42)
    assert subject.load_name == "world"


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            namespace="hello",
            version=42,
            parameters=LabwareDefinition2Parameters.model_construct(loadName="world"),  # type: ignore[call-arg]
            ordering=[],
            metadata=LabwareDefinitionMetadata.model_construct(  # type: ignore[call-arg]
                displayName="what a cool labware"
            ),
        )
    ],
)
def test_set_calibration_succeeds_in_ok_location(
    decoy: Decoy,
    subject: LabwareCore,
    mock_engine_client: EngineClient,
    labware_definition: LabwareDefinition2,
) -> None:
    """It should pass along an AddLabwareOffset if possible."""
    decoy.when(
        mock_engine_client.state.labware.get_definition_uri("cool-labware")
    ).then_return(
        uri_from_details(
            load_name=labware_definition.parameters.loadName,
            namespace=labware_definition.namespace,
            version=labware_definition.version,
        )
    )
    decoy.when(
        mock_engine_client.state.labware.get_display_name("cool-labware")
    ).then_return("what a cool labware")
    location = [
        OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="C2")
    ]
    decoy.when(
        mock_engine_client.state.geometry.get_offset_location("cool-labware")
    ).then_return(cast(LabwareOffsetLocationSequence, location))
    subject.set_calibration(Point(1, 2, 3))
    decoy.verify(
        mock_engine_client.add_labware_offset(
            LabwareOffsetCreate(
                definitionUri="hello/world/42",
                locationSequence=cast(LabwareOffsetLocationSequence, location),
                vector=LabwareOffsetVector(x=1, y=2, z=3),
            )
        ),
        mock_engine_client.execute_command(
            cmd.ReloadLabwareParams(
                labwareId="cool-labware",
            )
        ),
    )


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            namespace="hello",
            version=42,
            parameters=LabwareDefinition2Parameters.model_construct(loadName="world"),  # type: ignore[call-arg]
            ordering=[],
        )
    ],
)
def test_set_calibration_fails_in_bad_location(
    decoy: Decoy,
    subject: LabwareCore,
    mock_engine_client: EngineClient,
    labware_definition: LabwareDefinition2,
) -> None:
    """It should raise if you attempt to set calibration when the labware is not on deck."""
    decoy.when(
        mock_engine_client.state.labware.get_definition_uri("cool-labware")
    ).then_return(
        uri_from_details(
            load_name=labware_definition.parameters.loadName,
            namespace=labware_definition.namespace,
            version=labware_definition.version,
        )
    )
    decoy.when(
        mock_engine_client.state.geometry.get_offset_location("cool-labware")
    ).then_return(None)
    with pytest.raises(LabwareNotOnDeckError):
        subject.set_calibration(Point(1, 2, 3))


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            namespace="hello",
            parameters=LabwareDefinition2Parameters.model_construct(loadName="world"),  # type: ignore[call-arg]
            ordering=[],
            allowedRoles=[],
            stackingOffsetWithLabware={},
            stackingOffsetWithModule={},
            gripperOffsets={},
        )
    ],
)
def test_get_definition(subject: LabwareCore) -> None:
    """It should get the labware's definition as a dictionary."""
    assert subject.get_definition() == cast(
        LabwareDefDict,
        {
            "namespace": "hello",
            "parameters": {"loadName": "world"},
            "ordering": [],
            "allowedRoles": [],
            "stackingOffsetWithLabware": {},
            "stackingOffsetWithModule": {},
            "gripperOffsets": {},
        },
    )
    assert subject.get_parameters() == {  # type: ignore[comparison-overlap]
        "loadName": "world"
    }


def test_get_user_display_name(decoy: Decoy, mock_engine_client: EngineClient) -> None:
    """It should get the labware's user-provided label, if any."""
    decoy.when(
        mock_engine_client.state.labware.get_user_specified_display_name("cool-labware")
    ).then_return("Cool Label")

    subject = LabwareCore(labware_id="cool-labware", engine_client=mock_engine_client)
    result = subject.get_user_display_name()

    assert result == "Cool Label"
    assert result == subject.get_display_name()


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            ordering=[],
            metadata=LabwareDefinitionMetadata.model_construct(  # type: ignore[call-arg]
                displayName="Cool Display Name"
            ),
        )
    ],
)
def test_get_display_name(subject: LabwareCore) -> None:
    """It should get the labware's display name from the definition, if no label."""
    result = subject.get_display_name()

    assert result == "Cool Display Name"


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            parameters=LabwareDefinition2Parameters.model_construct(  # type: ignore[call-arg]
                loadName="load-name"
            ),
        ),
    ],
)
def test_get_name_load_name(subject: LabwareCore) -> None:
    """It should get the load name when no display name is defined."""
    result = subject.get_name()

    assert result == "load-name"


def test_get_name_display_name(decoy: Decoy, mock_engine_client: EngineClient) -> None:
    """It should get the user display name when one is defined."""
    decoy.when(
        mock_engine_client.state.labware.get_user_specified_display_name("cool-labware")
    ).then_return("my cool display name")

    subject = LabwareCore(labware_id="cool-labware", engine_client=mock_engine_client)

    result = subject.get_name()

    assert result == "my cool display name"


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            ordering=[],
            parameters=LabwareDefinition2Parameters.model_construct(isTiprack=True),  # type: ignore[call-arg]
        )
    ],
)
def test_is_tip_rack(subject: LabwareCore) -> None:
    """It should know whether it's a tip rack."""
    result = subject.is_tip_rack()

    assert result is True


@pytest.mark.parametrize(
    argnames=["labware_definition", "expected_result"],
    argvalues=[
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                ordering=[], allowedRoles=[LabwareRole.adapter]
            ),
            True,
        ),
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                ordering=[], allowedRoles=[LabwareRole.labware]
            ),
            False,
        ),
    ],
)
def test_is_adapter(expected_result: bool, subject: LabwareCore) -> None:
    """It should know whether it's an adapter."""
    result = subject.is_adapter()

    assert result is expected_result


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            ordering=[["A1", "B1"], ["A2", "B2"]],
        )
    ],
)
def test_get_well_columns(subject: LabwareCore) -> None:
    """It should get a well name columns from the definition."""
    result = subject.get_well_columns()

    assert result == [["A1", "B1"], ["A2", "B2"]]


def test_get_well_core(subject: LabwareCore) -> None:
    """It should get a well name grid from the definition."""
    result = subject.get_well_core("A1")

    assert isinstance(result, WellCore)
    assert result.get_name() == "A1"
    assert result.labware_id == "cool-labware"


def test_get_uri(
    decoy: Decoy, mock_engine_client: EngineClient, subject: LabwareCore
) -> None:
    """It should get a labware's URI from the core."""
    decoy.when(
        mock_engine_client.state.labware.get_definition_uri("cool-labware")
    ).then_return(LabwareUri("great/uri/42"))

    result = subject.get_uri()

    assert result == "great/uri/42"


def test_get_next_tip(
    decoy: Decoy, mock_engine_client: EngineClient, subject: LabwareCore
) -> None:
    """It should get the next available tip from the core."""
    decoy.when(
        mock_engine_client.state.tips.get_next_tip(
            labware_id="cool-labware",
            num_tips=8,
            starting_tip_name="B1",
            nozzle_map=None,
        )
    ).then_return("A2")

    starting_tip = WellCore(
        name="B1", labware_id="cool-labware", engine_client=mock_engine_client
    )
    result = subject.get_next_tip(
        num_tips=8, starting_tip=starting_tip, nozzle_map=None
    )

    assert result == "A2"


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            ordering=[],
            wells={
                "A1": RectangularWellDefinition2.model_construct(),  # type: ignore[call-arg]
                "H12": RectangularWellDefinition2.model_construct(),  # type: ignore[call-arg]
            },
            parameters=LabwareDefinition2Parameters.model_construct(isTiprack=True),  # type: ignore[call-arg]
        )
    ],
)
def test_reset_tips(
    decoy: Decoy, mock_engine_client: EngineClient, subject: LabwareCore
) -> None:
    """It should reset the tip state of a labware."""
    subject.reset_tips()
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.SetTipStateParams(
                labwareId="cool-labware",
                wellNames=["A1", "H12"],
                tipWellState=TipRackWellState.CLEAN,
            )
        )
    )


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            ordering=[],
            parameters=LabwareDefinition2Parameters.model_construct(isTiprack=False),  # type: ignore[call-arg]
            metadata=LabwareDefinitionMetadata.model_construct(  # type: ignore[call-arg]
                displayName="Cool Display Name"
            ),
        )
    ],
)
def test_reset_tips_raises_if_not_tip_rack(
    decoy: Decoy, mock_engine_client: EngineClient, subject: LabwareCore
) -> None:
    """It should raise an exception if the labware isn't a tip rack."""
    with pytest.raises(TypeError, match="Cool Display Name is not a tip rack"):
        subject.reset_tips()


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            ordering=[],
            wells={
                "A1": RectangularWellDefinition2.model_construct(),  # type: ignore[call-arg]
                "H12": RectangularWellDefinition2.model_construct(),  # type: ignore[call-arg]
            },
            parameters=LabwareDefinition2Parameters.model_construct(isTiprack=True),  # type: ignore[call-arg]
        )
    ],
)
def test_set_empty(
    decoy: Decoy, mock_engine_client: EngineClient, subject: LabwareCore
) -> None:
    """It should set the tip states to empty."""
    subject.set_empty()
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.SetTipStateParams(
                labwareId="cool-labware",
                wellNames=["A1", "H12"],
                tipWellState=TipRackWellState.EMPTY,
            )
        )
    )


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            ordering=[],
            parameters=LabwareDefinition2Parameters.model_construct(isTiprack=False),  # type: ignore[call-arg]
            metadata=LabwareDefinitionMetadata.model_construct(  # type: ignore[call-arg]
                displayName="Cool Display Name"
            ),
        )
    ],
)
def test_set_empty_raises_if_not_tip_rack(
    decoy: Decoy, mock_engine_client: EngineClient, subject: LabwareCore
) -> None:
    """It should raise an exception if the labware isn't a tip rack."""
    with pytest.raises(TypeError, match="Cool Display Name is not a tip rack"):
        subject.set_empty()


def test_get_tip_length(
    decoy: Decoy, mock_engine_client: EngineClient, subject: LabwareCore
) -> None:
    """It should get the tip length of a labware."""
    decoy.when(
        mock_engine_client.state.labware.get_tip_length("cool-labware")
    ).then_return(1.23)

    result = subject.get_tip_length()

    assert result == 1.23


def test_highest_z(
    decoy: Decoy, mock_engine_client: EngineClient, subject: LabwareCore
) -> None:
    """It should return the highest z of a labware."""
    decoy.when(
        mock_engine_client.state.geometry.get_labware_highest_z("cool-labware")
    ).then_return(9000.1)

    result = subject.highest_z

    assert result == 9000.1


def test_get_calibrated_offset(
    decoy: Decoy, mock_engine_client: EngineClient, subject: LabwareCore
) -> None:
    """It should get the calibrated offset."""
    decoy.when(
        mock_engine_client.state.geometry.get_labware_position("cool-labware")
    ).then_return(Point(1, 2, 3))

    result = subject.get_calibrated_offset()

    assert result == Point(1, 2, 3)


@pytest.mark.parametrize(
    "labware_definition",
    [
        LabwareDefinition2.model_construct(  # type: ignore[call-arg]
            ordering=[],
            parameters=LabwareDefinition2Parameters.model_construct(quirks=["quirk"]),  # type: ignore[call-arg]
        )
    ],
)
def test_get_quirks(subject: LabwareCore) -> None:
    """It should get a list of labware quirks."""
    result = subject.get_quirks()

    assert result == ["quirk"]


def test_get_deck_slot(
    decoy: Decoy, mock_engine_client: EngineClient, subject: LabwareCore
) -> None:
    """It should return its deck slot location, if in a slot."""
    decoy.when(
        mock_engine_client.state.geometry.get_ancestor_slot_name("cool-labware")
    ).then_return(DeckSlotName.SLOT_5)

    assert subject.get_deck_slot() == DeckSlotName.SLOT_5

    decoy.when(
        mock_engine_client.state.geometry.get_ancestor_slot_name("cool-labware")
    ).then_raise(LabwareNotOnDeckError("oh no"))

    assert subject.get_deck_slot() is None


def test_load_liquid(
    decoy: Decoy, mock_engine_client: EngineClient, subject: LabwareCore
) -> None:
    """It should pass loaded liquids to the engine."""
    mock_liquid = Liquid(
        _id="liquid-id", name="water", description=None, display_color=None
    )
    subject.load_liquid(volumes={"A1": 20, "B1": 30, "C1": 40}, liquid=mock_liquid)

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.LoadLiquidParams(
                labwareId="cool-labware",
                liquidId="liquid-id",
                volumeByWell={"A1": 20, "B1": 30, "C1": 40},
            )
        ),
        times=1,
    )


def test_load_empty(
    decoy: Decoy, mock_engine_client: EngineClient, subject: LabwareCore
) -> None:
    """It should pass empty liquids to the engine."""
    subject.load_empty(wells=["A1", "B1", "C1"])
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.LoadLiquidParams(
                labwareId="cool-labware",
                liquidId="EMPTY",
                volumeByWell={"A1": 0.0, "B1": 0.0, "C1": 0.0},
            )
        )
    )



"""Concurrency provider."""

import asyncio

import pytest

from opentrons.protocol_engine.resources import ConcurrencyProvider


@pytest.fixture
def subject() -> ConcurrencyProvider:
    """Creates a concurrency provider."""
    return ConcurrencyProvider()


def test_concurrency_provider_makes_locks(subject: ConcurrencyProvider) -> None:
    """Test concurrency provider makes locks."""
    lock1 = subject.lock_for_group("lock1")
    lock2 = subject.lock_for_group("lock2")
    assert isinstance(lock1, asyncio.Lock)
    assert isinstance(lock2, asyncio.Lock)
    assert lock1 is not lock2


def test_provides_locks_for_groups(subject: ConcurrencyProvider) -> None:
    """Test that lock returns the correct id."""
    lock1 = subject.lock_for_group("lock1")
    lock2 = subject.lock_for_group("lock1")
    assert lock1 is lock2
    assert isinstance(lock1, asyncio.Lock)



"""Tests for the APIs around deck conflicts during pipette movement."""

import pytest

from opentrons.protocol_api import ALL, COLUMN, ROW, SINGLE, ProtocolContext
from opentrons.protocol_api.core.engine.pipette_movement_conflict import (
    PartialTipMovementNotAllowedError,
)


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.16", "Flex")], indirect=True
)
def test_deck_conflicts_for_96_ch_a12_column_configuration(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should raise errors for the expected deck conflicts."""
    trash_labware = simulated_protocol_context.load_labware(
        "opentrons_1_trash_3200ml_fixed", "A3"
    )
    badly_placed_tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "C2"
    )
    well_placed_tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "C1"
    )
    tiprack_on_adapter = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul",
        "C3",
        adapter="opentrons_flex_96_tiprack_adapter",
    )

    thermocycler = simulated_protocol_context.load_module("thermocyclerModuleV2")
    tc_adjacent_plate = simulated_protocol_context.load_labware(
        "opentrons_96_wellplate_200ul_pcr_full_skirt", "A2"
    )
    accessible_plate = thermocycler.load_labware(
        "opentrons_96_wellplate_200ul_pcr_full_skirt"
    )

    instrument = simulated_protocol_context.load_instrument(
        "flex_96channel_1000", mount="left"
    )
    instrument.trash_container = trash_labware

    # ############  SHORT LABWARE  ################
    # These labware should be to the west of tall labware to avoid any partial tip deck conflicts
    badly_placed_labware = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "D2"
    )
    well_placed_labware = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "D3"
    )

    # ############ TALL LABWARE ##############
    simulated_protocol_context.load_labware(
        "opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical", "D1"
    )

    # ########### Use Partial Nozzles #############
    instrument.configure_nozzle_layout(style=COLUMN, start="A12")

    with pytest.raises(
        PartialTipMovementNotAllowedError, match="collision with items in deck slot"
    ):
        instrument.pick_up_tip(badly_placed_tiprack.wells_by_name()["A1"])

    instrument.pick_up_tip(well_placed_tiprack.wells_by_name()["A12"])
    instrument.aspirate(50, well_placed_labware.wells_by_name()["A4"])

    with pytest.raises(
        PartialTipMovementNotAllowedError, match="collision with items in deck slot D1"
    ):
        instrument.dispense(50, badly_placed_labware.wells()[0])

    with pytest.raises(
        PartialTipMovementNotAllowedError, match="collision with thermocycler lid"
    ):
        instrument.dispense(10, tc_adjacent_plate.wells_by_name()["A1"])

    instrument.dispense(10, tc_adjacent_plate.wells_by_name()["H2"])

    # No error cuz dispensing from high above plate, so it clears tuberack in west slot
    instrument.dispense(15, badly_placed_labware.wells_by_name()["A1"].top(150))

    thermocycler.open_lid()  # type: ignore[union-attr]

    with pytest.raises(
        PartialTipMovementNotAllowedError, match="outside of robot bounds"
    ):
        # Dispensing to A1 in an east-most slot using a configuration with column 12 would
        # result in a collision with the side of the robot.
        instrument.dispense(25, accessible_plate.wells_by_name()["A1"])

    instrument.drop_tip()

    # ######## CHANGE CONFIG TO ALL #########
    instrument.configure_nozzle_layout(style=ALL, tip_racks=[tiprack_on_adapter])

    # No error because of full config
    instrument.pick_up_tip()

    # No error NOW because of full config
    instrument.aspirate(50, badly_placed_labware.wells_by_name()["A1"])

    # No error
    instrument.dispense(50, accessible_plate.wells_by_name()["A1"])


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.20", "Flex")], indirect=True
)
def test_close_shave_deck_conflicts_for_96_ch_a12_column_configuration(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """Shouldn't raise errors for "almost collision"s."""
    res12 = simulated_protocol_context.load_labware("nest_12_reservoir_15ml", "C3")

    # Mag block and tiprack adapter are very close to the destination reservoir labware
    simulated_protocol_context.load_module("magneticBlockV1", "D2")
    simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_200ul",
        "B3",
        adapter="opentrons_flex_96_tiprack_adapter",
    )
    tiprack_8 = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_200ul", "B2"
    )
    hs = simulated_protocol_context.load_module("heaterShakerModuleV1", "C1")
    hs_adapter = hs.load_adapter("opentrons_96_deep_well_adapter")
    deepwell = hs_adapter.load_labware("nest_96_wellplate_2ml_deep")
    simulated_protocol_context.load_trash_bin("A3")
    p1000_96 = simulated_protocol_context.load_instrument("flex_96channel_1000")
    p1000_96.configure_nozzle_layout(style=SINGLE, start="A12", tip_racks=[tiprack_8])

    hs.close_labware_latch()  # type: ignore[union-attr]
    # Note
    p1000_96.distribute(
        15,
        res12["A6"],
        deepwell.columns()[6],
        disposal_volume=0,
    )


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.16", "Flex")], indirect=True
)
def test_deck_conflicts_for_96_ch_a1_column_configuration(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should raise errors for expected deck conflicts."""
    instrument = simulated_protocol_context.load_instrument(
        "flex_96channel_1000", mount="left"
    )
    trash_labware = simulated_protocol_context.load_labware(
        "opentrons_1_trash_3200ml_fixed", "A3"
    )
    instrument.trash_container = trash_labware

    badly_placed_tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "C2"
    )
    well_placed_tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "A1"
    )
    tiprack_on_adapter = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul",
        "C3",
        adapter="opentrons_flex_96_tiprack_adapter",
    )

    # ############  SHORT LABWARE  ################
    # These labware should be to the east of tall labware to avoid any partial tip deck conflicts
    badly_placed_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "B1"
    )
    well_placed_plate = simulated_protocol_context.load_labware(
        "nest_96_wellplate_200ul_flat", "B3"
    )

    # ############ TALL LABWARE ###############
    my_tuberack = simulated_protocol_context.load_labware(
        "opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical", "B2"
    )

    # ########### Use Partial Nozzles #############
    instrument.configure_nozzle_layout(style=COLUMN, start="A1")

    with pytest.raises(
        PartialTipMovementNotAllowedError, match="collision with items in deck slot"
    ):
        instrument.pick_up_tip(badly_placed_tiprack.wells_by_name()["H12"])

    # No error cuz within pipette extent bounds and no taller labware to right of tiprack
    instrument.pick_up_tip(well_placed_tiprack.wells_by_name()["A12"])

    # No error cuz no labware on right of plate, and also well A10 is juusst inside the right bound
    instrument.aspirate(25, well_placed_plate.wells_by_name()["A10"])

    # No error cuz dispensing from high above plate, so it clears tuberack on the right
    instrument.dispense(25, badly_placed_plate.wells_by_name()["A1"].top(150))

    with pytest.raises(
        PartialTipMovementNotAllowedError, match="collision with items in deck slot"
    ):
        instrument.aspirate(25, badly_placed_plate.wells_by_name()["A10"])

    with pytest.raises(
        PartialTipMovementNotAllowedError, match="outside of robot bounds"
    ):
        # Moving the 96 channel in column configuration with column 1
        # is incompatible with moving to a plate in B3 in the right most
        # column.
        instrument.aspirate(25, well_placed_plate.wells_by_name()["A11"])

    # No error because we're moving to column 1 of the plate with
    # column 1 of the 96 channel.
    instrument.aspirate(25, well_placed_plate.wells_by_name()["A1"])

    # No error cuz no taller labware on the right
    instrument.aspirate(10, my_tuberack.wells_by_name()["A1"])

    with pytest.raises(
        PartialTipMovementNotAllowedError, match="outside of robot bounds"
    ):
        # Raises error because drop tip alternation makes the pipette drop the tips
        # near the trash bin labware's right edge, which is out of bounds for column1 nozzles
        # We should probably move this tip drop location within the nozzles' accessible area,
        # but since we do not recommend loading the trash as labware (there are other things
        # wrong with that approach), it is not a critical issue.
        instrument.drop_tip()

    instrument.trash_container = None  # type: ignore
    simulated_protocol_context.load_trash_bin("C1")

    # This doesn't raise an error because it now treats the trash bin as an addressable area
    # and the bounds check doesn't yet check moves to addressable areas.
    # The aim is to do checks for ALL moves, but also, fix the offset used for tip drop alternation.
    instrument.drop_tip()

    # ######## CHANGE CONFIG TO ALL #########
    instrument.configure_nozzle_layout(style=ALL, tip_racks=[tiprack_on_adapter])

    # No error because of full config
    instrument.pick_up_tip()

    # No error NOW because of full config
    instrument.aspirate(50, badly_placed_plate.wells_by_name()["A1"])

    # No error NOW because of full config
    instrument.dispense(50, badly_placed_plate.wells_by_name()["A1"].bottom())


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "simulated_protocol_context", [("2.20", "Flex")], indirect=True
)
def test_deck_conflicts_for_96_ch_and_reservoirs(
    simulated_protocol_context: ProtocolContext,
) -> None:
    """It should raise errors for expected deck conflicts when moving to reservoirs.

    This test checks that the critical point of the pipette is taken into account,
    specifically when it differs from the primary nozzle.
    """
    instrument = simulated_protocol_context.load_instrument(
        "flex_96channel_1000", mount="left"
    )
    # trash_labware = protocol.load_labware("opentrons_1_trash_3200ml_fixed", "A3")
    # instrument.trash_container = trash_labware

    simulated_protocol_context.load_trash_bin("A3")
    right_tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "C3"
    )
    front_tiprack = simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul", "D2"
    )
    # Tall deck item in B3
    simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul",
        "B3",
        adapter="opentrons_flex_96_tiprack_adapter",
    )
    # Tall deck item in B1
    simulated_protocol_context.load_labware(
        "opentrons_flex_96_tiprack_50ul",
        "B1",
        adapter="opentrons_flex_96_tiprack_adapter",
    )

    # ############  RESERVOIRS  ################
    # These labware should be to the east of tall labware to avoid any partial tip deck conflicts
    reservoir_1_well = simulated_protocol_context.load_labware(
        "nest_1_reservoir_195ml", "C2"
    )
    reservoir_12_well = simulated_protocol_context.load_labware(
        "nest_12_reservoir_15ml", "B2"
    )

    # ########### Use COLUMN A1 Config #############
    instrument.configure_nozzle_layout(style=COLUMN, start="A1")

    instrument.pick_up_tip(front_tiprack.wells_by_name()["A12"])

    with pytest.raises(
        PartialTipMovementNotAllowedError, match="collision with items in deck slot"
    ):
        instrument.aspirate(10, reservoir_1_well.wells()[0])

    instrument.aspirate(25, reservoir_12_well.wells()[0])
    instrument.dispense(10, reservoir_12_well.wells()[1])

    with pytest.raises(
        PartialTipMovementNotAllowedError, match="collision with items in deck slot"
    ):
        instrument.dispense(15, reservoir_12_well.wells()[3])

    instrument.drop_tip()
    front_tiprack.reset()

    # ########### Use COLUMN A12 Config #############
    instrument.configure_nozzle_layout(style=COLUMN, start="A12")

    instrument.pick_up_tip(front_tiprack.wells_by_name()["A1"])
    instrument.aspirate(50, reservoir_1_well.wells()[0])
    with pytest.raises(
        PartialTipMovementNotAllowedError, match="collision with items in deck slot"
    ):
        instrument.dispense(10, reservoir_12_well.wells()[8])

    instrument.dispense(15, reservoir_12_well.wells()[11])
    instrument.dispense(10, reservoir_1_well.wells()[0])

    instrument.drop_tip()
    front_tiprack.reset()

    # ######## CHANGE CONFIG TO ROW H1 #########
    instrument.configure_nozzle_layout(style=ROW, start="H1", tip_racks=[front_tiprack])
    with pytest.raises(
        PartialTipMovementNotAllowedError, match="collision with items in deck slot"
    ):
        instrument.pick_up_tip(right_tiprack.wells_by_name()["A1"])

    instrument.pick_up_tip()
    instrument.aspirate(25, reservoir_1_well.wells()[0])

    instrument.drop_tip()
    front_tiprack.reset()

    # ######## CHANGE CONFIG TO ROW A1 #########
    instrument.configure_nozzle_layout(style=ROW, start="A1", tip_racks=[front_tiprack])

    with pytest.raises(
        PartialTipMovementNotAllowedError, match="outside of robot bounds"
    ):
        instrument.pick_up_tip()
    instrument.pick_up_tip(right_tiprack.wells_by_name()["H1"])

    with pytest.raises(
        PartialTipMovementNotAllowedError, match="collision with items in deck slot"
    ):
        instrument.aspirate(25, reservoir_1_well.wells()[0])

    instrument.drop_tip()
    front_tiprack.reset()



"""Test dispense commands."""

from datetime import datetime

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import (
    PipetteOverpressureError,
    StallOrCollisionDetectedError,
)

from opentrons.protocol_engine import (
    DeckPoint,
    LiquidHandlingWellLocation,
    WellOffset,
    WellOrigin,
)
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.dispense import (
    DispenseImplementation,
    DispenseParams,
    DispenseResult,
)
from opentrons.protocol_engine.commands.movement_common import StallOrCollisionError
from opentrons.protocol_engine.commands.pipetting_common import OverpressureError
from opentrons.protocol_engine.execution import MovementHandler, PipettingHandler
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import LabwareWellId
from opentrons.types import Point


@pytest.fixture
def subject(
    state_view: StateView,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    model_utils: ModelUtils,
) -> DispenseImplementation:
    """Get the implementation subject."""
    return DispenseImplementation(
        state_view=state_view,
        movement=movement,
        pipetting=pipetting,
        model_utils=model_utils,
    )


async def test_dispense_implementation(
    decoy: Decoy,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    subject: DispenseImplementation,
    state_view: StateView,
) -> None:
    """It should move to the target location and then dispense."""
    well_location = LiquidHandlingWellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )

    data = DispenseParams(
        pipetteId="pipette-id-abc123",
        labwareId="labware-id-abc123",
        wellName="A3",
        wellLocation=well_location,
        volume=50,
        flowRate=1.23,
    )

    decoy.when(
        await movement.move_to_well(
            pipette_id="pipette-id-abc123",
            labware_id="labware-id-abc123",
            well_name="A3",
            well_location=well_location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=50.0,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=1, y=2, z=3))

    decoy.when(
        state_view.geometry.get_nozzles_per_well(
            labware_id="labware-id-abc123",
            target_well_name="A3",
            pipette_id="pipette-id-abc123",
        )
    ).then_return(2)

    decoy.when(
        state_view.geometry.get_wells_covered_by_pipette_with_active_well(
            "labware-id-abc123", "A3", "pipette-id-abc123"
        )
    ).then_return(["A3", "A4"])

    decoy.when(
        state_view.pipettes.get_aspirated_volume("pipette-id-abc123")
    ).then_return(100)

    decoy.when(pipetting.get_state_view()).then_return(state_view)

    decoy.when(
        await pipetting.dispense_in_place(
            pipette_id="pipette-id-abc123",
            volume=50,
            flow_rate=1.23,
            push_out=None,
            is_full_dispense=False,
            correction_volume=0,
        )
    ).then_return(42)
    decoy.when(
        state_view.pipettes.get_liquid_dispensed_by_ejecting_volume(
            pipette_id="pipette-id-abc123", volume=42
        )
    ).then_return(34)

    result = await subject.execute(data)

    assert result == SuccessData(
        public=DispenseResult(volume=42, position=DeckPoint(x=1, y=2, z=3)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id-abc123",
                new_location=LabwareWellId(
                    labware_id="labware-id-abc123",
                    well_name="A3",
                ),
                new_deck_point=DeckPoint.model_construct(x=1, y=2, z=3),
            ),
            liquid_operated=update_types.LiquidOperatedUpdate(
                labware_id="labware-id-abc123",
                well_names=["A3", "A4"],
                volume_added=68,
            ),
            pipette_aspirated_fluid=update_types.PipetteEjectedFluidUpdate(
                pipette_id="pipette-id-abc123", volume=42
            ),
            ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                pipette_id="pipette-id-abc123", ready_to_aspirate=True
            ),
        ),
    )


async def test_overpressure_error(
    decoy: Decoy,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    subject: DispenseImplementation,
    model_utils: ModelUtils,
    state_view: StateView,
) -> None:
    """It should return an overpressure error if the hardware API indicates that."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    well_location = LiquidHandlingWellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )

    position = Point(x=1, y=2, z=3)

    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)

    data = DispenseParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location,
        volume=50,
        flowRate=1.23,
    )

    decoy.when(
        state_view.geometry.get_nozzles_per_well(
            labware_id=labware_id,
            target_well_name=well_name,
            pipette_id=pipette_id,
        )
    ).then_return(2)

    decoy.when(
        state_view.geometry.get_wells_covered_by_pipette_with_active_well(
            labware_id, well_name, pipette_id
        )
    ).then_return(["A3", "A4"])

    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=50.0,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_return(position)

    decoy.when(pipetting.get_state_view()).then_return(state_view)

    decoy.when(
        await pipetting.dispense_in_place(
            pipette_id=pipette_id,
            volume=50,
            flow_rate=1.23,
            push_out=None,
            is_full_dispense=False,
            correction_volume=0,
        ),
    ).then_raise(PipetteOverpressureError())

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)

    result = await subject.execute(data)

    assert result == DefinedErrorData(
        public=OverpressureError.model_construct(
            id=error_id,
            createdAt=error_timestamp,
            wrappedErrors=[matchers.Anything()],
            errorInfo={"retryLocation": (position.x, position.y, position.z)},
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id",
                new_location=LabwareWellId(
                    labware_id="labware-id",
                    well_name="well-name",
                ),
                new_deck_point=DeckPoint.model_construct(x=1, y=2, z=3),
            ),
            liquid_operated=update_types.LiquidOperatedUpdate(
                labware_id="labware-id",
                well_names=["A3", "A4"],
                volume_added=update_types.CLEAR,
            ),
            pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                pipette_id="pipette-id"
            ),
            ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                pipette_id="pipette-id", ready_to_aspirate=False
            ),
        ),
        state_update_if_false_positive=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id",
                new_location=LabwareWellId(
                    labware_id="labware-id",
                    well_name="well-name",
                ),
                new_deck_point=DeckPoint.model_construct(x=1, y=2, z=3),
            ),
        ),
    )


async def test_stall_error(
    decoy: Decoy,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    subject: DispenseImplementation,
    model_utils: ModelUtils,
    state_view: StateView,
) -> None:
    """It should return a stall error if the hardware API indicates that."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    well_location = LiquidHandlingWellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )

    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)

    data = DispenseParams(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location,
        volume=50,
        flowRate=1.23,
    )

    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=50.0,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_raise(StallOrCollisionDetectedError())

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)

    result = await subject.execute(data)

    assert result == DefinedErrorData(
        public=StallOrCollisionError.model_construct(
            id=error_id,
            createdAt=error_timestamp,
            wrappedErrors=[matchers.Anything()],
        ),
        state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
    )



"""Pipetting execution handler."""

from contextlib import nullcontext as does_not_raise
from typing import ContextManager, Dict, Optional, OrderedDict

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import (
    CommandParameterLimitViolated,
    CommandPreconditionViolated,
)
from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition,
    LabwareDefinition2,
)
from opentrons_shared_data.pipette.pipette_definition import ValidNozzleMaps

from opentrons.hardware_control import API as HardwareAPI
from opentrons.hardware_control.nozzle_manager import NozzleMap
from opentrons.hardware_control.protocols.types import FlexRobotType, OT2RobotType
from opentrons.hardware_control.types import (
    InstrumentProbeType,
    TipScrapeType,
    TipStateType,
)
from opentrons.protocol_engine.errors.exceptions import TipNotAttachedError
from opentrons.protocol_engine.execution.tip_handler import (
    HardwareTipHandler,
    VirtualTipHandler,
    create_tip_handler,
)
from opentrons.protocol_engine.resources import LabwareDataProvider
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import TipGeometry, TipPresenceStatus
from opentrons.types import Mount, MountType, NozzleConfigurationType, Point


@pytest.fixture
def mock_hardware_api(decoy: Decoy) -> HardwareAPI:
    """Get a mock in the shape of a HardwareAPI."""
    mock = decoy.mock(cls=HardwareAPI)
    decoy.when(mock.get_robot_type()).then_return(OT2RobotType)
    return mock


@pytest.fixture
def mock_state_view(decoy: Decoy) -> StateView:
    """Get a mock in the shape of a StateStore."""
    return decoy.mock(cls=StateView)


@pytest.fixture
def mock_nozzle_map(decoy: Decoy) -> NozzleMap:
    """Get a mock in the shape of a NozzleMap."""
    return decoy.mock(cls=NozzleMap)


@pytest.fixture
def mock_labware_data_provider(decoy: Decoy) -> LabwareDataProvider:
    """Get a mock LabwareDataProvider."""
    return decoy.mock(cls=LabwareDataProvider)


@pytest.fixture
def tip_rack_definition() -> LabwareDefinition:
    """Get a tip rack defintion value object."""
    return LabwareDefinition2.model_construct(namespace="test", version=42)  # type: ignore[call-arg]


MOCK_MAP = NozzleMap.build(
    physical_nozzles=OrderedDict({"A1": Point(0, 0, 0)}),
    physical_rows=OrderedDict({"A": ["A1"]}),
    physical_columns=OrderedDict({"1": ["A1"]}),
    starting_nozzle="A1",
    back_left_nozzle="A1",
    front_right_nozzle="A1",
    valid_nozzle_maps=ValidNozzleMaps(maps={"Full": ["A1"]}),
)


async def test_create_tip_handler(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
) -> None:
    """It should return virtual or real tip handlers depending on config."""
    decoy.when(mock_state_view.config.use_virtual_pipettes).then_return(False)
    assert isinstance(
        create_tip_handler(
            state_view=mock_state_view,
            hardware_api=mock_hardware_api,
        ),
        HardwareTipHandler,
    )

    decoy.when(mock_state_view.config.use_virtual_pipettes).then_return(True)
    assert isinstance(
        create_tip_handler(
            state_view=mock_state_view,
            hardware_api=mock_hardware_api,
        ),
        VirtualTipHandler,
    )


@pytest.mark.parametrize("tip_state", [TipStateType.PRESENT, TipStateType.ABSENT])
async def test_flex_pick_up_tip_state(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_labware_data_provider: LabwareDataProvider,
    tip_rack_definition: LabwareDefinition,
    tip_state: TipStateType,
    mock_hardware_api: HardwareAPI,
) -> None:
    """Test the protocol engine's pick_up_tip logic."""
    subject = HardwareTipHandler(
        state_view=mock_state_view,
        hardware_api=mock_hardware_api,
        labware_data_provider=mock_labware_data_provider,
    )
    decoy.when(mock_state_view.pipettes.get_mount("pipette-id")).then_return(
        MountType.LEFT
    )
    decoy.when(mock_state_view.pipettes.get_channels("pipette-id")).then_return(1)
    decoy.when(mock_state_view.pipettes.get_serial_number("pipette-id")).then_return(
        "pipette-serial"
    )
    decoy.when(mock_state_view.labware.get_definition("labware-id")).then_return(
        tip_rack_definition
    )
    decoy.when(
        mock_state_view.pipettes.get_nozzle_configuration("pipette-id")
    ).then_return(MOCK_MAP)
    decoy.when(
        mock_state_view.geometry.get_nominal_tip_geometry(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="B2",
        )
    ).then_return(TipGeometry(length=50, diameter=5, volume=300))

    decoy.when(
        await mock_labware_data_provider.get_calibrated_tip_length(
            pipette_serial="pipette-serial",
            labware_definition=tip_rack_definition,
            nominal_fallback=50,
        )
    ).then_return(42)

    if tip_state == TipStateType.PRESENT:
        await subject.pick_up_tip(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="B2",
        )
        decoy.verify(mock_hardware_api.cache_tip(Mount.LEFT, 42), times=1)
    else:
        decoy.when(
            await subject.verify_tip_presence(  # type: ignore[func-returns-value]
                pipette_id="pipette-id", expected=TipPresenceStatus.PRESENT
            )
        ).then_raise(TipNotAttachedError())
        # if a TipNotAttchedError is caught, we should not add any tip information
        with pytest.raises(TipNotAttachedError):
            await subject.pick_up_tip(
                pipette_id="pipette-id",
                labware_id="labware-id",
                well_name="B2",
            )
        decoy.verify(
            mock_hardware_api.cache_tip(
                mount=matchers.Anything(),
                tip_length=matchers.Anything(),
            ),
            ignore_extra_args=True,
            times=0,
        )


async def test_pick_up_tip(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    mock_labware_data_provider: LabwareDataProvider,
    tip_rack_definition: LabwareDefinition,
) -> None:
    """It should use the hardware API to pick up a tip."""
    subject = HardwareTipHandler(
        state_view=mock_state_view,
        hardware_api=mock_hardware_api,
        labware_data_provider=mock_labware_data_provider,
    )

    decoy.when(mock_state_view.labware.get_definition("labware-id")).then_return(
        tip_rack_definition
    )
    decoy.when(mock_state_view.pipettes.get_channels("pipette-id")).then_return(1)

    decoy.when(mock_state_view.pipettes.get_serial_number("pipette-id")).then_return(
        "pipette-serial"
    )

    decoy.when(mock_state_view.pipettes.get_mount("pipette-id")).then_return(
        MountType.LEFT
    )

    decoy.when(
        mock_state_view.pipettes.get_nozzle_configuration(pipette_id="pipette-id")
    ).then_return(MOCK_MAP)

    decoy.when(
        mock_state_view.geometry.get_nominal_tip_geometry(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="B2",
        )
    ).then_return(TipGeometry(length=50, diameter=5, volume=300))

    decoy.when(
        await mock_labware_data_provider.get_calibrated_tip_length(
            pipette_serial="pipette-serial",
            labware_definition=tip_rack_definition,
            nominal_fallback=50,
        )
    ).then_return(42)

    result = await subject.pick_up_tip(
        pipette_id="pipette-id",
        labware_id="labware-id",
        well_name="B2",
    )

    assert result == TipGeometry(length=42, diameter=5, volume=300)

    decoy.verify(
        await mock_hardware_api.tip_pickup_moves(
            mount=Mount.LEFT,
            presses=None,
            increment=None,
        ),
        mock_hardware_api.set_current_tiprack_diameter(
            mount=Mount.LEFT,
            tiprack_diameter=5,
        ),
        mock_hardware_api.set_working_volume(mount=Mount.LEFT, tip_volume=300),
    )


# todo(mm, 2024-10-17): Test that when verify_tip_presence raises,
# the hardware API state is NOT updated.
async def test_drop_tip(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    mock_labware_data_provider: LabwareDataProvider,
) -> None:
    """It should use the hardware API to drop a tip."""
    subject = HardwareTipHandler(
        state_view=mock_state_view,
        hardware_api=mock_hardware_api,
        labware_data_provider=mock_labware_data_provider,
    )

    decoy.when(mock_state_view.pipettes.get_mount("pipette-id")).then_return(
        MountType.RIGHT
    )
    decoy.when(
        mock_state_view.pipettes.get_nozzle_configuration("pipette-id")
    ).then_return(MOCK_MAP)

    await subject.drop_tip(pipette_id="pipette-id", home_after=True)

    decoy.verify(
        await mock_hardware_api.tip_drop_moves(
            mount=Mount.RIGHT,
            ignore_plunger=False,
            home_after=True,
            scrape_type=TipScrapeType.NONE,
        )
    )
    decoy.verify(mock_hardware_api.remove_tip(mount=Mount.RIGHT))
    decoy.verify(
        mock_hardware_api.set_current_tiprack_diameter(
            mount=Mount.RIGHT, tiprack_diameter=0
        )
    )


def test_add_tip(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    mock_labware_data_provider: LabwareDataProvider,
) -> None:
    """It should add a tip manually to the hardware API."""
    subject = HardwareTipHandler(
        state_view=mock_state_view,
        hardware_api=mock_hardware_api,
        labware_data_provider=mock_labware_data_provider,
    )

    tip = TipGeometry(
        length=50,
        diameter=5,
        volume=300,
    )

    decoy.when(mock_state_view.pipettes.get_mount("pipette-id")).then_return(
        MountType.LEFT
    )

    subject.cache_tip(pipette_id="pipette-id", tip=tip)

    decoy.verify(
        mock_hardware_api.cache_tip(mount=Mount.LEFT, tip_length=50),
        mock_hardware_api.set_current_tiprack_diameter(
            mount=Mount.LEFT,
            tiprack_diameter=5,
        ),
        mock_hardware_api.set_working_volume(mount=Mount.LEFT, tip_volume=300),
    )


def test_remove_tip(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    mock_labware_data_provider: LabwareDataProvider,
) -> None:
    """It should remove a tip manually from the hardware API."""
    subject = HardwareTipHandler(
        state_view=mock_state_view,
        hardware_api=mock_hardware_api,
        labware_data_provider=mock_labware_data_provider,
    )

    decoy.when(mock_state_view.pipettes.get_mount("pipette-id")).then_return(
        MountType.LEFT
    )

    subject.remove_tip(pipette_id="pipette-id")

    decoy.verify(
        mock_hardware_api.remove_tip(Mount.LEFT),
        mock_hardware_api.set_current_tiprack_diameter(Mount.LEFT, 0),
    )


@pytest.mark.parametrize(
    argnames=[
        "test_channels",
        "style",
        "primary_nozzle",
        "front_nozzle",
        "back_nozzle",
        "exception",
        "expected_result",
        "tip_result",
    ],
    argvalues=[
        [
            8,
            "COLUMN",
            "A1",
            None,
            None,
            does_not_raise(),
            {
                "primary_nozzle": "A1",
                "front_right_nozzle": "H1",
                "back_left_nozzle": "A1",
            },
            None,
        ],
        [
            8,
            "ROW",
            "A1",
            None,
            None,
            pytest.raises(CommandParameterLimitViolated),
            None,
            None,
        ],
        [
            8,
            "SINGLE",
            "A1",
            None,
            None,
            does_not_raise(),
            {"primary_nozzle": "A1"},
            None,
        ],
        [
            1,
            "SINGLE",
            "A1",
            None,
            None,
            pytest.raises(CommandPreconditionViolated),
            None,
            None,
        ],
        [
            8,
            "COLUMN",
            "A1",
            None,
            None,
            pytest.raises(CommandPreconditionViolated),
            None,
            TipGeometry(length=50, diameter=5, volume=300),
        ],
    ],
)
async def test_available_nozzle_layout(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    mock_labware_data_provider: LabwareDataProvider,
    test_channels: int,
    style: str,
    primary_nozzle: Optional[str],
    front_nozzle: Optional[str],
    back_nozzle: Optional[str],
    exception: ContextManager[None],
    expected_result: Optional[Dict[str, str]],
    tip_result: Optional[TipGeometry],
) -> None:
    """The virtual and hardware pipettes should return the same data and error at the same time."""
    hw_subject = HardwareTipHandler(
        state_view=mock_state_view,
        hardware_api=mock_hardware_api,
        labware_data_provider=mock_labware_data_provider,
    )
    virtual_subject = VirtualTipHandler(state_view=mock_state_view)
    decoy.when(mock_state_view.pipettes.get_channels("pipette-id")).then_return(
        test_channels
    )
    decoy.when(mock_state_view.pipettes.get_attached_tip("pipette-id")).then_return(
        tip_result
    )

    with exception:
        hw_result = await hw_subject.available_for_nozzle_layout(
            "pipette-id", style, primary_nozzle, front_nozzle, back_nozzle
        )
        virtual_result = await virtual_subject.available_for_nozzle_layout(
            "pipette-id", style, primary_nozzle, front_nozzle, back_nozzle
        )
        assert hw_result == virtual_result == expected_result


async def test_virtual_pick_up_tip(
    decoy: Decoy,
    mock_state_view: StateView,
    tip_rack_definition: LabwareDefinition,
) -> None:
    """It should use a virtual pipette to pick up a tip."""
    subject = VirtualTipHandler(state_view=mock_state_view)

    decoy.when(mock_state_view.labware.get_definition("labware-id")).then_return(
        tip_rack_definition
    )

    decoy.when(mock_state_view.pipettes.get_serial_number("pipette-id")).then_return(
        "pipette-serial"
    )

    decoy.when(
        mock_state_view.geometry.get_nominal_tip_geometry(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="B2",
        )
    ).then_return(TipGeometry(length=50, diameter=5, volume=300))

    result = await subject.pick_up_tip(
        pipette_id="pipette-id",
        labware_id="labware-id",
        well_name="B2",
    )

    assert result == TipGeometry(length=50, diameter=5, volume=300)

    decoy.verify(
        mock_state_view.pipettes.validate_tip_state("pipette-id", False),
        times=1,
    )


async def test_virtual_drop_tip(decoy: Decoy, mock_state_view: StateView) -> None:
    """It should use a virtual pipette to drop a tip."""
    subject = VirtualTipHandler(state_view=mock_state_view)

    await subject.drop_tip(pipette_id="pipette-id", home_after=None)

    decoy.verify(
        mock_state_view.pipettes.validate_tip_state("pipette-id", True),
        times=1,
    )


async def test_get_tip_presence_on_ot2(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_hardware_api: HardwareAPI,
    mock_labware_data_provider: LabwareDataProvider,
) -> None:
    """It should use the hardware API to  up a tip."""
    subject = HardwareTipHandler(
        state_view=mock_state_view,
        hardware_api=mock_hardware_api,
        labware_data_provider=mock_labware_data_provider,
    )

    result = await subject.get_tip_presence(pipette_id="pipette-id")
    assert result == TipPresenceStatus.UNKNOWN


@pytest.mark.parametrize("hw_tip_state", [TipStateType.ABSENT, TipStateType.PRESENT])
async def test_get_tip_presence_on_ot3(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_labware_data_provider: LabwareDataProvider,
    hw_tip_state: TipStateType,
) -> None:
    """It should use the hardware API to  up a tip."""
    try:
        from opentrons.hardware_control.ot3api import OT3API

        ot3_hardware_api = decoy.mock(cls=OT3API)
        decoy.when(ot3_hardware_api.get_robot_type()).then_return(FlexRobotType)

        subject = HardwareTipHandler(
            state_view=mock_state_view,
            hardware_api=ot3_hardware_api,
            labware_data_provider=mock_labware_data_provider,
        )

        decoy.when(mock_state_view.pipettes.get_mount("pipette-id")).then_return(
            MountType.LEFT
        )
        decoy.when(
            await ot3_hardware_api.get_tip_presence_status(Mount.LEFT)
        ).then_return(hw_tip_state)
        result = await subject.get_tip_presence(pipette_id="pipette-id")
        assert result == TipPresenceStatus.from_hw_state(hw_tip_state)

    except ImportError:
        pass


@pytest.mark.parametrize(
    "expected", [TipPresenceStatus.ABSENT, TipPresenceStatus.PRESENT]
)
async def test_verify_tip_presence_on_ot3(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_labware_data_provider: LabwareDataProvider,
    expected: TipPresenceStatus,
) -> None:
    """It should use the hardware API to  up a tip."""
    try:
        from opentrons.hardware_control.ot3api import OT3API

        ot3_hardware_api = decoy.mock(cls=OT3API)
        decoy.when(ot3_hardware_api.get_robot_type()).then_return(FlexRobotType)

        subject = HardwareTipHandler(
            state_view=mock_state_view,
            hardware_api=ot3_hardware_api,
            labware_data_provider=mock_labware_data_provider,
        )
        decoy.when(mock_state_view.pipettes.get_mount("pipette-id")).then_return(
            MountType.LEFT
        )

        decoy.when(
            mock_state_view.pipettes.get_nozzle_configuration("pipette-id")
        ).then_return(MOCK_MAP)

        await subject.verify_tip_presence("pipette-id", expected, None)

        decoy.verify(
            await ot3_hardware_api.verify_tip_presence(
                Mount.LEFT, expected.to_hw_state(), None
            )
        )

    except ImportError:
        pass


@pytest.mark.parametrize(
    argnames=[
        "pipette_channels",
        "style",
        "channels_in_nozzle_map",
        "back_left_nozzle",
        "front_right_nozzle",
        "expected_tip_presence_supported",
        "expected_follow_singular_sensor",
    ],
    argvalues=[
        [
            1,
            NozzleConfigurationType.SINGLE,
            1,
            "A1",
            "A1",
            True,
            None,
        ],
        [
            8,
            NozzleConfigurationType.COLUMN,
            8,
            "A1",
            "H1",
            True,
            None,
        ],
        [
            8,
            NozzleConfigurationType.SINGLE,
            1,
            "A1",
            "A1",
            False,
            None,
        ],
        [
            8,
            NozzleConfigurationType.COLUMN,
            3,
            "A1",
            "C1",
            False,
            None,
        ],
        [
            8,
            NozzleConfigurationType.COLUMN,
            4,
            "A1",
            "D1",
            True,
            None,
        ],
        [
            96,
            NozzleConfigurationType.SINGLE,
            1,
            "A1",
            "A1",
            False,
            None,
        ],
        [
            96,
            NozzleConfigurationType.COLUMN,
            8,
            "A1",
            "H1",
            True,
            InstrumentProbeType.PRIMARY,
        ],
        [
            96,
            NozzleConfigurationType.COLUMN,
            3,
            "A1",
            "C1",
            False,
            None,
        ],
        [
            96,
            NozzleConfigurationType.COLUMN,
            8,
            "A12",
            "H12",
            True,
            InstrumentProbeType.SECONDARY,
        ],
        [
            96,
            NozzleConfigurationType.ROW,
            12,
            "A1",
            "A12",
            True,
            None,
        ],
        [
            96,
            NozzleConfigurationType.ROW,
            8,
            "A5",
            "A12",
            True,
            InstrumentProbeType.SECONDARY,
        ],
        [
            96,
            NozzleConfigurationType.ROW,
            8,
            "A1",
            "A8",
            True,
            InstrumentProbeType.PRIMARY,
        ],
        [
            96,
            NozzleConfigurationType.ROW,
            3,
            "A1",
            "A3",
            False,
            None,
        ],
        [
            96,
            NozzleConfigurationType.ROW,
            3,
            "A10",
            "A12",
            False,
            None,
        ],
    ],
)
async def test_tip_presence_config(
    decoy: Decoy,
    mock_state_view: StateView,
    mock_nozzle_map: NozzleMap,
    mock_labware_data_provider: LabwareDataProvider,
    pipette_channels: int,
    style: NozzleConfigurationType,
    channels_in_nozzle_map: int,
    back_left_nozzle: str,
    front_right_nozzle: str,
    expected_tip_presence_supported: bool,
    expected_follow_singular_sensor: Optional[InstrumentProbeType],
) -> None:
    """It should check for tip presence when supported and use only one sensor in certain configs."""
    try:
        from opentrons.hardware_control.ot3api import OT3API

        ot3_hardware_api = decoy.mock(cls=OT3API)
        decoy.when(ot3_hardware_api.get_robot_type()).then_return(FlexRobotType)

        subject = HardwareTipHandler(
            state_view=mock_state_view,
            hardware_api=ot3_hardware_api,
            labware_data_provider=mock_labware_data_provider,
        )
        decoy.when(mock_state_view.pipettes.get_mount("pipette-id")).then_return(
            MountType.LEFT
        )

        decoy.when(mock_state_view.pipettes.get_channels("pipette-id")).then_return(
            pipette_channels
        )
        decoy.when(mock_nozzle_map.tip_count).then_return(channels_in_nozzle_map)
        decoy.when(mock_nozzle_map.configuration).then_return(style)
        decoy.when(mock_nozzle_map.back_left).then_return(back_left_nozzle)
        decoy.when(mock_nozzle_map.front_right).then_return(front_right_nozzle)

        decoy.when(
            mock_state_view.pipettes.get_nozzle_configuration("pipette-id")
        ).then_return(mock_nozzle_map)
        (
            tip_presence_supported,
            follow_singular_sensor,
        ) = subject.get_tip_presence_config("pipette-id")
        assert tip_presence_supported == expected_tip_presence_supported
        assert follow_singular_sensor == expected_follow_singular_sensor
    except ImportError:
        pass
